"""debmdk.py -- A Module to Generate Points on a 'DEBMDK' Problem
    
    This module provides functions to generate points on an m-dimensional 
    Knee problem ('DEBMDK') [1]_.

    Copyright (C) 2016
    Computational Optimization and Innovation (COIN) Laboratory
    Department of Computer Science and Engineering
    Michigan State University
    428 S. Shaw Lane, Engineering Building
    East Lansing, MI 48824-1226, USA


    References
    ----------
    .. [1] Branke J., Deb K., Dierolf H., Osswald M. (2004) Finding Knees in 
        Multi-objective Optimization. In: Yao X. et al. (eds) Parallel Problem 
        Solving from Nature - PPSN VIII. PPSN 2004. Lecture Notes in Computer 
        Science, vol 3242. Springer, Berlin, Heidelberg.

.. moduleauthor:: AKM Khaled Talukder <talukde1@msu.edu>

"""

import numpy as np
from viz.generators import dtlz2

__all__ = ["surface"]

def rf(x, k=1):
    r"""The radius function for the knee problem.
    """
    return 5.0 + (10.0 * (x - 0.5) * (x - 0.5)) \
            + ((2.0 / k) * np.cos(2.0 * k * np.pi * x))

def surface(r=1, k=1, n=10, m=2, mode='lhc', **kwargs):
    r"""Generate `n` number of points on an `m`-dimensional knee ('DEBMDK') problem [1]_.

    The Pareto-optimal front for "DEBMDK" problem is basically a hypersurface of
    a distorted 'm'-sphere, such that there are 'K' number of bulges on it.

    The radius of the `m`-sphere is specified as 1. The total number of "knees"
    is specified in `k`. The point generation is currently done in two ways of 
    random sampling -- Latin Hypercube (LHC) and LHC with normalization. Other 
    ways will be added later.

    For non-random and equidistant point samplin,g we use Das-Dennis's method 
    [3]_ to generate points using NBI technique.

    Parameters
    ----------
    k : int, optional
        The total number of knees. Default 1 when optional.
    n : int, optional
        The total number of points. Default 10 when optional.
    m : int, optional
        The dimension of the sphere. Default 2 when optional.
    mode : str, {'lhc', 'lhcl2', 'dd'}, optional
        If `mode = `lhc``, then LHC sampling will be used and points will be generated
        using standard spherical coordinate systems. If `mode = `lhcl2``, then we will
        use a normalized LHC sampling to generate uniformly distributed points on the
        sphere using the method described in [2]_. If 'mode = 'dd', then we will generate
        points using the subproblem decomposition technique used in NBI method 
        (a.k.a. "Das-Dennis's Approach") discussed in [3]_. Default first when optional.

    Other Parameters
    ----------------
    delta : float, optional
        `delta` value for normalized LHC, this is used so that we only keep vectors
        `V` such that `np.linalg.norm(V, 1) > delta`. The default value is 0.0001 but
        you might want to change it according to your application.

    Returns
    -------
    F : ndarray
        `n` points on the `m`-sphere, i.e. `|F| = n x m`.
    X : ndarray
        `n` points of the `m-1` dimensional spherical coordinate values, i.e. `|X| = n x (m-1)`.

    References
    ----------
    .. [1] Branke J., Deb K., Dierolf H., Osswald M. (2004) Finding Knees in 
        Multi-objective Optimization. In: Yao X. et al. (eds) Parallel Problem 
        Solving from Nature - PPSN VIII. PPSN 2004. Lecture Notes in Computer 
        Science, vol 3242. Springer, Berlin, Heidelberg.

    .. [2] Simon C., "Generating uniformly distributed numbers on a sphere," online: 
        http://corysimon.github.io/articles/uniformdistn-on-sphere/ 

    .. [3] I. Das and J. E. Dennis, "Normal-Boundary Intersection: A New Method for 
        Generating the Pareto Surface in Nonlinear Multicriteria Optimization Problems," 
        SIAM Journal on Optimization, vol. 8, (3), pp. 631-27, 1998.
    """
    # This is needed for dtlz2
    delta = kwargs['delta'] if (len(kwargs) > 0 and 'delta' in kwargs) else 0.0001
    
    F, X = dtlz2.surface(r=r, n=n, m=m, mode=mode, delta=delta)

    g = 1.0
    vrf = np.vectorize(rf)
    R = (np.sum(vrf(X), axis = 1) / X.shape[1])
    F = (F.T * (g * R)).T

    return F, X



"""dtlz2.py -- A Module to Generate Points on a 'DTLZ2' Problem ('m'-Sphere)
    
    This module provides functions to generate points on a 'DTLZ2' problem [1]_.

    Copyright (C) 2016
    Computational Optimization and Innovation (COIN) Laboratory
    Department of Computer Science and Engineering
    Michigan State University
    428 S. Shaw Lane, Engineering Building
    East Lansing, MI 48824-1226, USA

    References
    ----------
    .. [1] K. Deb, L. Thiele, M. Laumanns and E. Zitzler. Scalable Multi-Objective 
        Optimization Test Problems. CEC 2002, p. 825 - 830, IEEE Press, 2002.

.. moduleauthor:: AKM Khaled Talukder <talukde1@msu.edu>

"""

import numpy as np
from viz.utils import sampling as smp

__all__ = ["surface"]


def surface(r=1, n=10, m=2, mode='lhc', **kwargs):
    r"""Generate `n` number of points on 'DTLZ2' problem (`m`-shpere).
    
    This function generates points on a 'DTLZ2' problem [1]_, which is
    in fact an 'm'-sphere. Therefore, this function samples points on an
    'm'-sphere.

    The radius of the `m`-sphere is specified in `r`. The point generation is 
    currently done in two ways of random sampling -- Latin Hypercube (LHC) and
    LHC with normalization. Other ways will be added later.

    For non-random and equidistant point samplin,g we use Das-Dennis's method 
    [3]_ to generate points using NBI technique.

    Parameters
    ----------
    r : float, optional
        The radius of the sphere. Default 1 when optional.
    n : int, optional
        The total number of points. Default 10 when optional.
    m : int, optional
        The dimension of the sphere. Default 2 when optional.
    mode : str, {'lhc', 'lhcl2', 'dd'}, optional
        If `mode = `lhc``, then LHC sampling will be used and points will be generated
        using standard spherical coordinate systems. If `mode = `lhcl2``, then we will
        use a normalized LHC sampling to generate uniformly distributed points on the
        sphere using the method described in [2]_. If 'mode = 'dd', then we will generate
        points using the subproblem decomposition technique used in NBI method 
        (a.k.a. "Das-Dennis's Approach") discussed in [3]_. Default first when optional.

    Other Parameters
    ----------------
    delta : float, optional
        `delta` value for normalized LHC, this is used so that we only keep vectors
        `V` such that `np.linalg.norm(V, 1) > delta`. The default value is 0.0001 but
        you might want to change it according to your application.

    Returns
    -------
    F : ndarray
        `n` points on the `m`-sphere, i.e. `|F| = n x m`.
    X : ndarray
        `n` points of the `m-1` dimensional spherical coordinate values, i.e. `|X| = n x (m-1)`.

    References
    ----------
    .. [1] K. Deb, L. Thiele, M. Laumanns and E. Zitzler. Scalable Multi-Objective 
        Optimization Test Problems. CEC 2002, p. 825 - 830, IEEE Press, 2002.

    .. [2] Simon C., "Generating uniformly distributed numbers on a sphere," online: 
        http://corysimon.github.io/articles/uniformdistn-on-sphere/ 

    .. [3] I. Das and J. E. Dennis, "Normal-Boundary Intersection: A New Method for 
        Generating the Pareto Surface in Nonlinear Multicriteria Optimization Problems," 
        SIAM Journal on Optimization, vol. 8, (3), pp. 631-27, 1998.
    """

    mkeys = {'lhc', 'lhcl2', 'dd'}
    if mode not in mkeys:
        raise KeyError("'mode' has to be any of {:s}.".format(str(mkeys)))
    if m < 2:
        raise ValueError("\'m\' must be greater than 2.")

    delta = kwargs['delta'] if (len(kwargs) > 0 and 'delta' in kwargs) else 0.0001

    F = np.ones((n, m))
    if mode == 'lhc':
        theta = smp.lhc(n=n, m=m - 1)
        X = np.copy(theta)
         
        # # print for sanity check
        # for i in range(m):
        #     if i < m-1:
        #         s1 = "".join(["sin(t_{:d} * pi/2)".format(v) for v in range(0,i)])
        #         s2 = "cos(t_{:d} * pi/2)".format(i)
        #         s = s1 + s2
        #     else:
        #         s1 = "".join(["sin(t_{:d} * pi/2)".format(v) for v in range(0,i-1)])
        #         s2 = "sin(t_{:d} * pi/2)".format(i-1)
        #         s = s1 + s2
        #     print("x_{:d} = ".format(i) + s)
        
        # This is what is described in wikipedia
        for i in range(m):
            if i < m-1:
                F[:,i] = np.prod(np.sin(theta[:,0:i] * (np.pi / 2)), axis = 1) \
                        * np.cos(theta[:,i] * (np.pi / 2))
            else:
                F[:,i] = np.prod(np.sin(theta[:,0:i-1] * (np.pi / 2)), axis = 1) \
                        * np.sin(theta[:,i-1] * (np.pi / 2))

        # This is reverse of what is described in wikipedia
        # # print for sanity check
        # for i in range(m):
        #     s = "".join(["sin(t_{:d} * pi/2)".format(v) for v in range(0,m - (i + 1))])
        #     if i > 0:
        #         s = s + "cos(t_{:d} * pi/2)".format(m - (i + 1)) 
        #     print("x_{:d} = ".format(i) + s)
        # for i in range(m):
        #     F[:,i] = np.prod(np.sin(theta[:,0:m - (i + 1)] * (np.pi / 2)), axis = 1)
        #     if i > 0:
        #         F[:,i] = F[:,i] * np.cos(theta[:,m - (i + 1)] * (np.pi / 2)) 

    elif mode == 'lhcl2':
        Y = smp.lhcl2(n=n, m=m, delta=delta)
        Y_ = np.linalg.norm(Y, 2, axis=1)
        F = (Y.T / Y_).T
        X = np.zeros((n, m-1))
        
        # # Print for sanity check
        # for i in range(m-2):
        #     s = "arccos(x_{:d}".format(i) + " / " + "√(" \
        #             + " + ".join(["x_{:d}²".format(v) for v in range(m-1,i-1,-1)]) \
        #             + "))"
        #     print("t_{:d} = ".format(i) + s)
        #
        # s1 = "arccos(x_{:d}".format(m-2) + " / " + "√(" \
        #         + " + ".join(["x_{:d}²".format(v) for v in range(m-1,m-3,-1)]) \
        #         + "))" + " if x_{:d} >= 0".format(m-1)
        # s2 = "(2 * pi - arccos(x_{:d}".format(m-2) + " / " + "√(" \
        #         + " + ".join(["x_{:d}²".format(v) for v in range(m-1,m-3,-1)]) \
        #         + ")))" + " if x_{:d} < 0".format(m-1)
        # print("t_{:d} = ".format(m-2) + s1 + " or " + s2)

        # This is what is explained in wikipedia
        for i in range(m-2):
            denom = np.linalg.norm(F[:,i:m], 2, axis=1)
            Inz = np.nonzero(denom)
            X[Inz,i] = np.arccos(F[Inz,i] / denom[Inz]) / (np.pi / 2) 
        denom = np.linalg.norm(F[:,m-2:m], 2, axis=1)
        Inz = np.nonzero(denom)
        X[Inz,m-2] = np.arccos(F[Inz,m-2] / denom[Inz]) / (np.pi / 2)
        
        # We don't need this since already all F[:,-1] > 0 
        # Ip = np.where(F[:,-1] >= 0)[0]
        # if len(Ip) > 0:
        #     X[Ip, m-2] = np.arccos(F[Ip,m-2] / denom[Ip]) / (np.pi / 2)
        # In = np.where(F[:,-1] < 0)[0]
        # if len(In) > 0:
        #     X[In, m-2] = (2 * np.pi) - (np.arccos(F[In,m-2] / denom[In]) / (np.pi / 2))

        # This is reverse of what is explained in wikipedia
        # for i in range(m-1):
        #     denom = np.linalg.norm(F[:,(m-1)-i::-1], 2, axis = 1)
        #     X[:,i] = np.arccos(F[:,(m-1)-i] / denom) / (np.pi / 2)

    elif mode == 'dd':
        F = smp.das_dennis(n = n, m = m)
        X = np.zeros((F.shape[0], m-1))
        for i in range(m-2):
            denom = np.linalg.norm(F[:,i:m], 2, axis=1)
            Inz = np.nonzero(denom)
            X[Inz,i] = np.arccos(F[Inz,i] / denom[Inz]) / (np.pi / 2)
        
        denom = np.linalg.norm(F[:,m-2:m], 2, axis=1)
        Inz = np.nonzero(denom)
        X[Inz,m-2] = np.arccos(F[Inz,m-2] / denom[Inz]) / (np.pi / 2)
    
    # We flip it since all MOPs are flipped and we
    # are using the definitions used in wikipedia.
    F = F[:,::-1] 
    return r * F, X



"""dm.py -- A Collection of Different Utility Functions for Decision Making (DM)
    
    This module provides different utility functions for decision making (DM) in
    multi-objective optimization (MOP) scenario, for example, finding knee 
    points and the relative trade-off of the points in the objective function space.

    Copyright (C) 2016
    Computational Optimization and Innovation (COIN) Laboratory
    Department of Computer Science and Engineering
    Michigan State University
    428 S. Shaw Lane, Engineering Building
    East Lansing, MI 48824-1226, USA

.. moduleauthor:: AKM Khaled Talukder <talukde1@msu.edu>

"""

import numpy as np
from scipy.spatial import cKDTree
import viz.utils.transform as tr

__all__ = ["epsilons", "nadir", "ideal", "knees", "tradeoff"]

# A set of good epsilon values to be used while computing 
# the tradeoff neighborhoods with respect to the dimension 
# of the space.
epsilons = {'3d': 0.125, '4d': 0.125, '5d': 0.25, \
            '6d': 0.375, '7d': 0.497, '8d': 0.497, \
            '9d': 0.6, '10d': 0.7}

def nadir(F):
    r"""Find the Nadir objective vector from 'F'.

    Parameters
    ----------
    F : ndarray
        A sample of `n` data points in `m` dimension, i.e. `|F| = n x m`.

    Returns
    -------
    Fmax : ndarray
        The 'numpy.max' values from each column.
    """

    Fmax = np.max(F, axis=0)
    return Fmax

def ideal(F):
    r"""Find the Ideal objective vector from 'F'.
    
    Parameters
    ----------
    F : ndarray
        A sample of `n` data points in `m` dimension, i.e. `|F| = n x m`.

    Returns
    -------
    Fmin : ndarray
        The 'numpy.min' values from each column.
    """

    Fmin = np.min(F, axis=0)
    return Fmin

def knees(Mu):
    r"""Find the best trade-off values.

    This function filters out the best trade-off values from a vector of
    trade-off values in 'Mu'. This function considers the values greater
    or equal to 'mean(Mu) + t * stdev(Mu)' are the `best` trade-off 
    values, where t = 2.75 found to be optimal. It was found empirically.

    Parameters
    ----------
    Mu : ndarray
        A vector of 'n' floats, i.e. '|Mu| = n x 1'.

    Returns
    -------
    Ik : ndarray (or None)
        An array of integers, each value is the index of best trade-off 
        values in 'Mu'. A 'None' will be returned if 'Ik' is empty.
    """

    # filter out nan and inf values    
    Mu_ = Mu[(~np.isnan(Mu)) & (~np.isinf(Mu))]

    
    mu, sigma = Mu_.mean(), Mu_.std()
    # We are considering 2*sigma are the "best" 
    # trade-off points, i.e. knee points.
    dev = (Mu_ - mu) / sigma
    
    # Ik = np.where(dev >= 2)[0]
    
    d = mu + 2 * sigma
    Ik = np.where(Mu_ >= d)[0]

    if Ik.shape[0] < 1 and dev.max() > 1:
        Ik = np.array([np.array(np.argmax(Mu_))])

    return Ik if Ik.shape[0] > 0 else None


def tradeoff(F, epsilon=0.125, k=None, penalize_extremes=False):
    r"""Calculate the trade-off values in the objective vectors 'F' to find the 'knee points'.

    Calculate the trade-off weight 'mu(.)' of each points in the objective vectors in 'F'. 
    The method is described in [1]_.

    By default, the neighbourhood will be considered as `epsilon = 0.125` neighbour, 
    therefore, all the objective function vector will be normalized within '[0.0, 1.0]' 
    before the calculation.
    
    Parameters
    ----------
    F : ndarray
        Input data points 'F' of 'm'-dimensional objective function vector, 
        i.e. '|F| = n x m'.
    epsilon : float, optional
        The 'epsilon' neighborhood radius. Default 0.125 when optional.
    k : int, optional
        The number of neighboring points. If 'k' is set, 'epsilon' will 
        not be used. Default 'None' when optional.
    penalize_extremes : boolean, optional
        The method discussed in [1]_ puts higher weights on extreme points. 
        If this variable is set to 'True', the extreme points will have lower 
        trade-off weights, leaving only the knee points in the non-extreme 
        (other/intermediary) part of the Pareto-optimal front. Default 'None' 
        when optional

    Returns
    -------
    Mu : ndarray
        An array of trade-off values of each corresponding vector in 'F'. 
        It's an array of floats, i.e. '|Mu| = n x 1'. `Mu` values are normalized 
        within [0.0,1.0].
    Ik : ndarray
        The indices of "knee" points in 'F'. Returns 'None' if no such point exists. 

    References
    ----------
    .. [1] Rachmawati, L. & Srinivasan, D., Multiobjective Evolutionary Algorithm 
        With Controllable Focus on the Knees of the Pareto Front. IEEE Transactions 
        on Evolutionary Computation, Institute of Electrical and Electronics 
        Engineers (IEEE), 2009, 13, 810-824
    """

    # use either epsilon or k
    if k:
        epsilon = None

    n,m = F.shape
    F_ = tr.normalize(F, lb=np.zeros(F.shape), ub=np.ones(F.shape))
    
    # Threshold on neighborhood size: 2m + 2
    s = 2 * m + 2
    # s = 2 * m + 1
    # s = 2 * m
    
    tree = cKDTree(F_)
    
    Mu = np.full(n, -np.inf)
    for i in range(n):
        # use epsilon
        if epsilon:
            S = np.array(tree.query_ball_point(F_[i], r = epsilon))
        # use k
        else:
            S = tree.query(F_[i], k = k)[1]
        
        # if neighborhood is sparse, get at least 2m + 2 number of points
        # The total number of vertices in a m-dim simplex is m+1
        if S.shape[0] < s:
            S = tree.query(F_[i], k = s + 1)[1]

        diff = F_[S] - F_[i]
        loss = np.maximum(0, diff).sum(axis = 1)
        gain = np.maximum(0, -diff).sum(axis = 1)

        np.warnings.filterwarnings('ignore')
        T = loss / gain
        
        Mu[i] = np.nanmin(T)
        
        if penalize_extremes:
            denom = max(F_[i]) - min(F_[i])
            Mu[i] = np.nanmin(T)/denom if denom > 0.0 else Mu[i]
        
    Ik = knees(Mu)
    Mu = tr.normalize(Mu, lb=np.array([0.0]), ub=np.array([1.0]))
    return Mu, Ik



"""sampling.py -- A Collection of Different Utility Functions for DOE
    
    This module provides different utility functions for random sampling, 
    for example, Latin Hypercube Sampling (LHS) etc.

    Copyright (C) 2016
    Computational Optimization and Innovation (COIN) Laboratory
    Department of Computer Science and Engineering
    Michigan State University
    428 S. Shaw Lane, Engineering Building
    East Lansing, MI 48824-1226, USA

.. moduleauthor:: AKM Khaled Talukder <talukde1@msu.edu>

"""

import random
import warnings
import numpy as np
from scipy.special import comb

__all__ = ["grid", "lhs", "lhcl2", "das_dennis", "risez"]

def grid(n=100, m=2):
    r""" A simple meshgrid generator for 'm'-dimensional coordinate.

    A very simple function to generate grid points in an 'm'-dimensional
    coordinate. It will generate `⌈n^(1/m)⌉^m' number of points.

    Parameters
    ----------
    n : int, optional
        The number of points. Default is 100 when optional.
    m : int
        The number of dimensions. Default is 2 when optional.

    Returns
    -------
    F : ndarray
        A sample of `⌈n^(1/m)⌉^m` data points in `m` dimension, 
        i.e. `|F| = ⌈n^(1/m)⌉^m x m`.

    """
    d = np.ceil(n ** (1/m)).astype(np.int64)
    if m > 1:
        M = np.meshgrid(*[np.linspace(0,1,d) for _ in range(m)])
        F = np.hstack(M).swapaxes(0,1).reshape(m,-1).T
    else:
        F = np.linspace(0,1,d)[:,None] # to make it (n,) to (n,1)
    if F.shape[0] != n:
        warnings.warn("Not possible to generate {:d} points in grid.".format(n))
        warnings.warn("Genenrated {:d} points instead.".format(F.shape[0]))
    return F


def lhc(n=10, m=2):
    r""" Latin Hyper-cube Sampling (LHS) of `n` points in `m` dimension.

    A very simple LHS code, every point is within `[0.0, 1.0]`.

    Parameters
    ----------
    n : int, optional
        The number of points. Default is 10 when optional.
    m : int
        The number of dimensions. Default is 2 when optional.

    Returns
    -------
    F : ndarray
        A sample of `n` data points in `m` dimension, i.e. `|F| = n x m`.

    """
    d = 1.0 / float(n) ;
    N = np.arange(n)
    F = np.zeros((n, m))
    for i in range(m):
        F[:,i] = (N * d) + (((N + 1.0) * d) - (N * d)) * np.random.random(n)
        np.random.shuffle(F[:,i])
    return F


def lhcl2(n=10, m=2, delta=0.0001):
    r""" Latin Hyper-cube Sampling of `n` points in `m` dimension with L2-norm constraint.

    Latin hypercube sampling n samples of m-dimensional points. This function guarantees 
    that 2-norm of each sample is greater than 0.0001. This function is slower than `lhc()`.
    Every point is within `[0.0, 1.0]`.

    Parameters
    ----------
    n : int
        The number of points.
    m : int
        The number of dimensions.

    Returns
    -------
    F : ndarray
        A sample of `n` data points in `m` dimension, i.e. `|F| = n x m`.

    """
    F = np.zeros((n, m))
    i = 0
    # skip_count = 0
    while i < n:
        k = n - i
        T = lhc(k, m)
        idx = np.argwhere(np.linalg.norm(T, 2, axis = 1) > delta)
        # skip_count += k - idx.shape[0]
        j = i + idx.shape[0]
        if j < n:
            F[i:j,:] = T[0:idx.shape[0],:]
            i = j
        else:
            F[i:,:] = T[0:k,:]
            i += k
    # print('skip_count =', skip_count)
    return F


def _das_dennis_inner(ref_dirs, ref_dir, p, beta, depth):
    r""" The inner function for '_das_dennis()'.

    The recursive inner function part for the _das_dennis() procedure.
    """
    if depth == len(ref_dir) - 1:
        ref_dir[depth] = beta / (1.0 * p)
        ref_dirs.append(ref_dir[None, :])
    else:
        for i in range(beta + 1):
            ref_dir[depth] = 1.0 * i / (1.0 * p)
            _das_dennis_inner(ref_dirs, np.copy(ref_dir), p, beta - i, depth + 1)


def _das_dennis(p, m):
    r"""Finding weights for the subproblem decomposition in NBI. 
    
    This function will be called by 'das_dennis()'. This function will
    recursively create the decomposition weights depending on the 'p'
    and 'm' values.
    """
    if p == 0:
        return np.full((1, m), 1 / m)
    else:
        ref_dirs = []
        ref_dir = np.full(m, np.nan)
        _das_dennis_inner(ref_dirs, ref_dir, p, p, 0)
        return np.concatenate(ref_dirs, axis = 0)


def das_dennis(n=100, m=2, manifold='sphere'):
    r""" Generating equidistant points on an 'm'-simplex (or 'm'-sphere).

    This function uses the recursive subdivision method described by 
    the NBI approach in [1]_ to generate equidistant points on an 
    'm'-dimensional simplex or on an 'm'-sphere.

    The method described in [1]_ is known as "Das-Dennis's" method.
    This function will approximate the partition parameter 'p' from given
    'n' and 'm' values. Therefore, if 'n' is not equal to '((m + p - 1)!)/(p!(m-1)!)',
    this function will try to generate points close to that number.

    Parameters
    ----------
    n : int, optional
        The number of points. Default is 100 when optional.
    m : int
        The number of dimensions. Default is 2 when optional.
    manifold : str, {'sphere', 'simplex'}, optional
        If the 'manifold' is 'simplex', the function will generate points 
        on an 'm'-simplex. If the 'manifold' is 'sphere', they will be on an 
        'm'-sphere. First when default.

    Returns
    -------
    F : ndarray
        A sample of `((m + p - 1)!)/(p!(m-1)!)` data points in `m` dimension, 
        i.e. `|F| = ((m + p - 1)!)/(p!(m-1)!) x m`.

    References
    ----------

    .. [1] I. Das and J. E. Dennis, "Normal-Boundary Intersection: A New Method for 
        Generating the Pareto Surface in Nonlinear Multicriteria Optimization Problems," 
        SIAM Journal on Optimization, vol. 8, (3), pp. 631-27, 1998.
    """
    mkeys = {'sphere', 'simplex'}
    if manifold not in mkeys:
        raise KeyError("Invalid 'manifold', use any of {:s}".format(str(mkeys)))

    if m == 1:
        d = np.ceil(n ** (1/m)).astype(np.int64)
        return np.linspace(0,1,d)[:,None]

    p = 1
    n_ = comb(m + p - 1, p, exact=False).astype(np.int64)
    while n_ <= n:
        p += 1
        n_ = comb(m + p - 1, p, exact=False).astype(np.int64)
    p = p - 1
    
    R = _das_dennis(m=m, p=p)
    if R.shape[0] != n:
        warnings.warn("Das-Dennis's method couldn't generate {:d} points.".format(n))
        warnings.warn("Genenrated {:d} points instead.".format(R.shape[0]))
    
    if manifold == 'sphere':
        F = R / np.linalg.norm(R, axis=1)[:,None]
    else:
        F = R
    return F



"""transform.py -- A collection of different utility functions for different data transformations
    
    This module provides different utility functions for 
    the transformation of data and other things.

    Copyright (C) 2016
    Computational Optimization and Innovation (COIN) Laboratory
    Department of Computer Science and Engineering
    Michigan State University
    428 S. Shaw Lane, Engineering Building
    East Lansing, MI 48824-1226, USA

.. moduleauthor:: AKM Khaled Talukder <talukde1@msu.edu>

"""

import numpy as np


__all__ = ["normalize", "pfindices"] 


def normalize(A, lb=None, ub=None):
    r"""Normalize a matrix within `[lb, ub]`.

    Just a simple normalization procedure that uses `numpy`.

    Parameters
    ----------
    A : array_like
        The input matrix.
    lb : array_like, optional
        The lower bound of each column. Default 'None' when optional.
    ub : array_like, optional
        The upper bound of each column. Default `None` when optional.

    Returns
    -------
    B : ndarray
        The result after normalizing the input array `A`.

    """
    if lb is None:
        lb = 0  if len(A.shape) == 1 else np.zeros(A.shape[1])
    if ub is None:
        ub = 1 if len(A.shape) == 1 else np.ones(A.shape[1])
    num = (A - A.min(axis=0)) * (ub - lb)
    denom = A.max(axis=0) - A.min(axis=0)
    if isinstance(denom, np.ndarray):
        denom[denom == 0] = 1
    else:
        denom = 1 if denom == 0.0 else denom
    B = lb + num / denom
    return B


def pfindices(A):
    r"""Find indices of non-dominated points in `A`.

    Find the indices of all non-dominated points (Pareto-optimal front)
    of the data points in matrix 'A'.

    Parameters
    ----------
    A : array_like
        The input matrix.

    Returns
    -------
    I : ndarray
        The indices of all non-dominated points the input array `A`.

    """

    # Testing code
    # G = np.array([[2, 1], [1, 4], [6, 2], [2, 6]])
    # fig = plt.figure()
    # ax = fig.gca()
    # ax.scatter(G[:,0], G[:,1])
    # ax.set_xlim((0,7))
    # ax.set_ylim((0,7))
    # plt.show()
    # i = tr.pfindices(G)
    # print(i)

    # G = np.array([[2.0, 1.0], [1.0, 4.0], [4.0, 1.0], \
    #               [6.0, 0.5], [1.0, 5.0], [0.5, 6.0], \
    #               [2.0, 3.0], [1.5, 4.0], [3.0, 2.0], [4.0, 4.0]])
    # fig = plt.figure()
    # ax = fig.gca()
    # ax.scatter(G[:,0], G[:,1])
    # ax.set_xlim((0,7))
    # ax.set_ylim((0,7))
    # i = tr.pfindices(G)
    # print(i)

    D = np.zeros((A.shape[0], A.shape[0]), dtype=np.bool)
    for i in range(A.shape[1]):
        x = A[:,i]
        if i > 0:
            D = D & (x[..., np.newaxis] > x[np.newaxis, ...])
        else:
            D = x[..., np.newaxis] > x[np.newaxis, ...]
    # print(D)
    I = np.where(~D.any(axis=1))[0]
    return I



"""paletteviz.py -- A customized and more flexible Paletteviz plotting module. 

    This module provides a customized and more flexible function for PaletteViz [1]_ 
    plotting. This module also provides different relevant fucntions, parameters and tools.

    Copyright (C) 2016
    Computational Optimization and Innovation (COIN) Laboratory
    Department of Computer Science and Engineering
    Michigan State University
    428 S. Shaw Lane, Engineering Building
    East Lansing, MI 48824-1226, USA
    
    References
    ----------
    .. [1] A. K. A. Talukder and K. Deb, "PaletteViz: A Visualization Method for Functional 
        Understanding of High-Dimensional Pareto-Optimal Data-Sets to Aid Multi-Criteria 
        Decision Making," in IEEE Computational Intelligence Magazine, vol. 15, no. 2, 
        pp. 36-48, May 2020, doi: 10.1109/MCI.2020.2976184.

.. moduleauthor:: AKM Khaled Talukder <talukde1@msu.edu>

"""
import os
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import matplotlib.colors as mc
from matplotlib.colors import ListedColormap
from mpl_toolkits.mplot3d import Axes3D
from viz.plotting.radviz import get_radviz_coordinates
from viz.plotting.star import get_star_coordinates
from viz.plotting.utils import set_polar_anchors, set_polar_anchor_labels, pop, group_labels_by_appearance
from viz.tda import simple_shape
from viz.utils import io


__all__ = ["get_palette_star_coordinates", "get_palette_radviz_coordinates", "plot", \
        "camera_angles_star", "camera_angles_radviz"]


# Some good camera angles for paletteviz plots.
camera_angles_star = {
    'dtlz2': {'3d': (20,30), '4d':(-40,20), '8d': (-25,30)}, \
    'dtlz2-nbi': {'3d': (-60,30), '4d':(-65,20), '5d': (-60,30), '8d': (-60,30)}, \
    'debmdk': {'3d': (130,20), '4d': (140,20), '8d': (155,20)}, \
    'debmdk-nbi': {'3d': (100,25), '4d': (0,20), '8d': (-45,20)}, \
    'debmdk-all': {'3d': (135,20), '4d': (160,20), '8d': (165,25)}, \
    'debmdk-all-nbi': {'3d': (85,20), '4d': (10,20), '8d': (-35,20)}, \
    'dtlz8': {'3d': (0,15), '4d': (-5,10), '6d': (-180,20), '8d': (-180,25)}, \
    'dtlz8-nbi': {'3d': (0,15), '4d': (-5,10), '6d': (-180,20), '8d': (-180,25)}, \
    'c2dtlz2': {'3d': (-125,20), '4d': (125,20), '5d': (60,30), '8d': (-20,20)}, \
    'c2dtlz2-nbi': {'3d': (30,30), '4d': (-20,15), '5d': (95,35), '8d': (-20,20)}, \
    'cdebmdk': {'3d': (-60,35), '4d': (70,30), '8d': (80,30)}, \
    'cdebmdk-nbi': {'3d': (-70,45), '4d': (60,30), '8d': (80,30)}, \
    'c0dtlz2': {'3d': (120,30), '4d': (165,30), '8d': (-150,45)}, \
    'c0dtlz2-nbi': {'3d': (125,20), '4d': (165,20), '8d': (160,35)}, \
    'crash-nbi': {'3d': (0,10)}, 'crash-c1-nbi': {'3d': (-25,15)}, 'crash-c2-nbi': {'3d': (-15,15)}, \
    'gaa': {'10d': (-15,25)}, \
    'gaa-nbi': {'10d': (0,25)}
}

camera_angles_radviz = {
    'dtlz2': {'3d': (-50,30), '4d':(-55,25), '8d': (-50,15)}, \
    'dtlz2-nbi': {'3d': (-50,30), '4d':(-65,20), '5d': (-60,30), '8d': (-60,30)}, \
    'debmdk': {'3d': (-35,25), '4d': (-30,25), '8d': (-40,15)}, \
    'debmdk-nbi': {'3d': (-65,25), '4d': (-60,20), '8d': (-55,-20)}, \
    'debmdk-all': {'3d': (100,20), '4d': (-30,20), '8d': (-20,15)}, \
    'debmdk-all-nbi': {'3d': (-75,25), '4d': (-70,20), '8d': (-145,20)}, \
    'dtlz8': {'3d': (-60,30), '4d': (-20,20), '6d': (0,35), '8d': (5,35)}, \
    'dtlz8-nbi': {'3d': (-60,30), '4d': (-20,20), '6d': (30,15), '8d': (55,25)}, \
    'c2dtlz2': {'3d': (-60,35), '4d': (125,20), '5d': (35,25), '8d': (-20,20)}, \
    'c2dtlz2-nbi': {'3d': (85,25), '4d': (160,20), '5d': (35,25), '8d': (-20,20)}, \
    'cdebmdk': {'3d': (-165,25), '4d': (-10,30), '8d': (-120,25)}, \
    'cdebmdk-nbi': {'3d': (180,30), '4d': (-115,30), '8d': (-80,25)}, \
    'c0dtlz2': {'3d': (180,25), '4d': (20,20), '8d': (-115,35)}, \
    'c0dtlz2-nbi': {'3d': (175,25), '4d': (20,20), '8d': (160,35)}, \
    'crash-nbi': {'3d': (-175,30)}, 'crash-c1-nbi': {'3d': (100,25)}, 'crash-c2-nbi': {'3d': (-100,25)}, \
    'gaa': {'10d': (0,25)}, \
    'gaa-nbi': {'10d': (20,40)}
}

def make_partitions(P, K, B, L, n_partitions):
    r"""Merge depth contours to make total of n_partitions number of layers.

    This function will be used by `get_palette_star_coordinates()` and 
    `get_palette_radviz_coordinates()` functions. Here we try to merge 
    the individual depth contours so that the final number of layers are 
    at most `n_partitions`. Let's say there are total 20 depth contours, 
    and if the user wants to visualise at most 4 layers, then we merge 
    every 5 depth contours from the start to make the final number of 
    layers to be 4. If `n_partitions` is bigger than the number of depth 
    contours, the total number of final layers will be the same as the 
    total number of depth contours.

    Parameters
    ----------
    P : ndarray 
        `n` number of 2D points found from radviz or star-coordinates.
    K : array_like, 2D
        The coordinate of the 2D anchor points found from radviz or star-coordinates.
    B : array_like, 2D
        The bounds of the anchor positions found from radviz or star-coordinates.
    L : ndarray, jagged
        The indices of each depth contours. The top row has the lowest
        depth and the bottom row has the highest depth. Each column is
        the index of the points in the original space.
    n_partitions : int
        The total number of layers in the final PaletteViz plot. We recommend 4.
        Default `float('inf')` when optional. When default, the total number of 
        layers in the PaletteViz will be the same as the number of depth contours. 
        Also if `n_partitions` is bigger than the total number depth contours, 
        the total number of layers in the PaletteViz will be equal to the total
        number of depth contours.

    Returns
    -------
    P : ndarray of float
        An ndarray of PaletteViz coordinates (i.e. `|P| = n x 3`).
    K : ndarray of float
        The positions of the anchor points (i.e. `|K| = m x 2`),
    B : ndarray of float
        The lower bound and upper bound of `P`. 
    Z : ndarray of float 
        An ndarray of z-coordinate values of all the anchor-sets 
        (i.e. `|Z| = n_partitions x 1`). `K` and `Z` values will be used to draw 
        anchor points and the polygon. 
    """
    n,m,p = P.shape[0], P.shape[1], L.shape[0]
    n_partitions = p if n_partitions >= p else n_partitions
    # q = number of layers in each partition
    # r = number of layers left after dividing them into n_partition layers
    # dz = the gap in z-axis between each pair of consedutive layers
    q, r, dz = p // n_partitions, p % n_partitions, 1 / n_partitions
    P_ = np.zeros((n, 3))
    z, Z = 1.0, np.zeros(n_partitions)
    for j,i in enumerate(range(0, p-r, q)):
        L_ = L[i:i+q]
        for l in L_:
            Id = l.astype(int)
            P_[Id,0:m] = P[Id,:]
            P_[Id,m] = np.ones(Id.shape[0]) * z
        Z[j] = z
        z = z - dz
    # if there is any remaining layer, merge them with the last one
    if r > 0:
        for i in range(L.shape[0]-1,(L.shape[0]-r)-1,-1):
            Id = L[i].astype(int)
            P_[Id,0:m] = P[Id,:]
            P_[Id,m] = np.ones(Id.shape[0]) * Z[-1]
    else:
        # if the last layer has very small number of points, smaller than
        # 1/10-th of the number of points in the previous layer, merge them 
        # with the previous layer. Therefore, the final number of layers 
        # will be n_partitions-1.
        if Z.shape[0] > 1:
            Iz = np.where(P_[:,-1] == Z[-1])[0]
            Iz_1 = np.where(P_[:,-1] == Z[-2])[0]
            if Iz.shape[0] < Iz_1.shape[0] / 10:
                P_[Iz,-1] = Z[-2]
                Z = Z[:-1]
    # update P
    P = P_
    # add z-bounds to B
    B[0] = np.append(B[0], np.min(Z))
    B[1] = np.append(B[1], np.max(Z))
    return (P,K,B,Z)


def get_palette_star_coordinates(X, depth_contours=None, n_partitions=float('inf'), \
                                    inverted=True, normalized=True, project_collapse=True, \
                                    verbose=False):
    r"""Generate Star-coordinates from data points `X`.

    Maps all the data points in `X` (i.e. `|X| = n x m`) onto 
    star-coordinate [1]_ positions. 

    Parameters
    ----------
    X : ndarray 
        `n` number of `m` dimensiomal points as input.
    depth_contours : ndarray or str path, optional
        An `ndarray` containing the depth contours or a path to the file 
        containing depth contour indices. Default 'None' when optional.
    n_partitions : int, optional
        The total number of layers in the final PaletteViz plot. We recommend 4.
        Default `float('inf')` when optional. When default, the total number of 
        layers in the PaletteViz will be the same as the number of depth contours. 
        Also if `n_partitions` is bigger than the total number depth contours, 
        the total number of layers in the PaletteViz will be equal to the total
        number of depth contours.
    inverted : bool, optional
        See `viz.plotting.star` for more details.
    normalized : bool, optional
        See `viz.plotting.star` for details.
    project_collapse : bool, optional
        See `viz.tda.simple_shape` module for more details.
    verbose : bool, optional
        Verbose level. Default `False` when optional.

    Returns
    -------
    P : ndarray of float
        An ndarray of PaletteViz coordinates (i.e. `|P| = n x 3`).
    K : ndarray of float
        The positions of the anchor points (i.e. `|K| = m x 2`),
    B : ndarray of float
        The lower bound and upper bound of `P`. 
    Z : ndarray of float 
        An ndarray of z-coordinate values of all the anchor-sets 
        (i.e. `|Z| = n_partitions x 1`). `K` and `Z` values will be used to draw 
        anchor points and the polygon.
    """

    L = None
    if depth_contours is None:
        # compute layers
        if verbose:
            print("Computing depth contours since no file provided.")
        L = simple_shape.depth_contours(X, project_collapse=project_collapse, verbose=verbose) 
    else: 
        if isinstance(depth_contours, str):
            # load depth contours
            if verbose:
                print("Loading depth contours from {0:s}.".format(depth_contours))
            L = io.loadtxt(depth_contours, dtype=int, delimiter=',') 
        elif isinstance(depth_contours, np.ndarray):
            # use depth contours
            if verbose:
                print("Using depth contours ndarray.")
            L = depth_contours

    if L is not None:
        (P,K,B) = get_star_coordinates(X, inverted=inverted, normalized=normalized)
        (P,K,B,Z) = make_partitions(P,K,B,L,n_partitions)
        return (P, K, B, Z)
    else:
        raise ValueError("No depth contours found.")


def get_palette_radviz_coordinates(X, depth_contours=None, n_partitions=float('inf'), \
                                    spread_factor='auto', normalized=True, project_collapse=True, \
                                    verbose=False):
    r"""Generate Radviz coordinates from data points `X`.

    Maps all the data points in `X` (i.e. `|X| = n x m`) onto 
    Radviz coordinate [1]_ positions. 

    Parameters
    ----------
    X : ndarray 
        `n` number of `m` dimensiomal points as input.
    depth_contours : ndarray or str path, optional
        An `ndarray` containing the depth contours or a path to the file 
        containing depth contour indices. Default 'None' when optional.
    n_partitions : int, optional
        The total number of layers in the final PaletteViz plot. We recommend 4.
        Default `float('inf')` when optional. When default, the total number of 
        layers in the PaletteViz will be the same as the number of depth contours. 
        Also if `n_partitions` is bigger than the total number depth contours, 
        the total number of layers in the PaletteViz will be equal to the total
        number of depth contours.
    spread_factor : str {'auto'} or float, optional
        See `viz.plotting.radviz` for more details.
    normalized : bool, optional
        See `viz.plotting.radviz` for details.
    project_collapse : bool, optional
        See `viz.tda.simple_shape` module for more details.
    verbose : bool, optional
        Verbose level. Default `False` when optional.

    Returns
    -------
    P : ndarray of float
        An ndarray of PaletteViz coordinates (i.e. `|P| = n x 3`).
    K : ndarray of float
        The positions of the anchor points (i.e. `|K| = m x 2`),
    B : ndarray of float
        The lower bound and upper bound of `P`. 
    Z : ndarray of float 
        An ndarray of z-coordinate values of all the anchor-sets 
        (i.e. `|Z| = n_partitions x 1`). `K` and `Z` values will be used to draw 
        anchor points and the polygon.
    """

    L = None
    if depth_contours is None:
        # compute layers
        if verbose:
            print("Computing depth contours since no file provided.")
        L = simple_shape.depth_contours(X, project_collapse=project_collapse, verbose=verbose) 
    else: 
        if isinstance(depth_contours, str):
            # load depth contours
            if verbose:
                print("Loading depth contours from {0:s}.".format(depth_contours))
            L = io.loadtxt(depth_contours, dtype=int, delimiter=',') 
        elif isinstance(depth_contours, np.ndarray):
            # use depth contours
            if verbose:
                print("Using depth contours ndarray.")
            L = depth_contours

    if L is not None:
        (P,K,B) = get_radviz_coordinates(X, spread_factor=spread_factor, normalized=normalized)
        (P,K,B,Z) = make_partitions(P,K,B,L,n_partitions)
        return (P, K, B, Z)
    else:
        raise ValueError("No depth contours found.")


def plot(A, ax=None, depth_contours=None, mode='star', n_partitions=float('inf'), \
        s=1.0, c=mc.TABLEAU_COLORS['tab:blue'], draw_axes=False, \
        draw_anchors={'labels': [0,1,2,3], 'polygons': [0,1,2,3], 'circles': [0,1,2,3]}, **kwargs):
    r"""A customized and more enhanced PaletteViz plot.

    This PaletteViz plot is customized for the experiments. It allows both
    Radviz and Star-coodinate based PaletteViz plots, depending on the options
    selected.
    
    Parameters
    ----------
    A : ndarray 
        `n` number of `m` dim. points to be plotted.
    ax : An `mpl_toolkits.mplot3d.axes.Axes3D` object, optional
        Default `None` when optional.
    depth_contours : ndarray or str path, optional
        See `get_palette_star_coordinates()` or `get_palette_radviz_coordinates()` 
        functions for detail. Default 'None' when optional.
    mode : str {'star', 'radviz'}, optional
        By which way each layer will be mapped. It can be Radviz, Star-coordinate etc.
        If this value is 'star', then Star-coordinate plot will be used. If the value
        is 'radviz', then Radviz plot will be used. Default 'star' when optional.
    n_partitions : int, optional
        See `get_palette_star_coordinates()` or `get_palette_radviz_coordinates()`
        functions for details. Default `float('inf')` when optional.
    s : float or 1-D array_like, optional
        Point size, or an array of point sizes. Default 1.0 when optional.
    c : A `matplotlib.colors` object, str or an array RGBA color values.
        Colors to be used. Default `mc.TABLEAU_COLORS['tab:blue']` when 
        optional.
    draw_axes : bool, optional
        If `True`, the radviz plot will show axes. Default `False` when optional.
    draw_anchors : None or {'labels': list of int, 'polygons': list of int, 'circles': list of int}, optional
        If `None`, there will be no polygons, circles or the labels. Each `list` of `int` 
        denotes which layer will have polygon, labels or circles. The `labels`, `polygons` and/or 
        `circles` will be drawn depending on the `list` of `int` provided. Default `{'labels': [0,1,2,3], `
        `'polygons': [0,1,2,3], 'circles': [0,1,2,3]}` when optional. 

    Other Parameters
    ----------------
    euler : tuple (i.e. a pair) of int, optional
        The azmiuth and elevation angle. Default `(-60,30)` when optional.
    title : str, optional
        The plot title. Default `None` when optional.
    labels : str, array_like or list of str, optional
        A string or an array/list of strings for labeling each point. Which basically
        means the class label of each row. Default `None` when optional. This will be
        used to set the legend in the figure. If `None` there will be no legend.
    spread_factor : str {'auto'} or float, optional
        See `viz.plotting.radviz` for more details.
    inverted : bool, optional
        See `viz.plotting.star` for more details.
    normalized : bool, optional
        See `viz.plotting.star` for details.
    project_collapse : bool, optional
        See `viz.tda.simple_shape` module for more details.
    colorbar : (Cbc, Cbg, Cbl, Cbp) a tuple of (ndarray, ndarray, str, float), optional
        If a user wants to put a colorbar, a tuple `(Cbc, Cbg, Cbl)` tuple can be 
        provided. `Cbc` is an array of RGBA color values or an `matplotlib.colors` 
        object. The gradient of the colorbar is specified in `Cbg` which is an 1-D 
        array of float. Cbl is the label of the colorbar, a string. `Cbp` is the 
        colorbar padding width in `float`. `colorbar` is default `None` when optional. 
    lims : 3-tuple of pairs of float
        The data point limits on three coordinates, each tuple is the limits on the
        coordinates to be used by `matplotlib.axes.Axes.set_x/y/zlim. If the `lims`
        is `(xl, yl, zl)`, then `xl`, `yl` and `zl` will be used for x-limit, y-limit
        and z-limit, respectively. Default `(None, None, None)` when optional.
    hide_layers : tuple of int
        List of layer indices to hide. Where there are L layers, The top layer is 
        indexed at 0 which will have z-axis value of 1.0 and the bottom layer is 
        indexed at L-1 which will have z-axis value of 0.0, when there are L layers. 
        Default `None` when optional.
    anchor_linewidth : float, optional
        See `set_polar_anchor()` function for details.
    anchor_label_prefix : str, optional
        See `set_anchor_labels()` function for details.
    anchor_label_fontsize : str or int, optional
        See `set_anchor_labels()` function for details.
    anchor_label_fontname : str, optional
        See `set_anchor_labels()` function for details.
    anchor_label_fontstyle : str, optional
        See `set_anchor_labels()` function for details.
    verbose : bool, optional
        The verbosity. Default 'False' when optional.
    **kwargs : dict
        All other keyword args for matplotlib `scatter()` function.

    Returns
    -------
    ax : `mpl_toolkits.mplot3d.axes.Axes3D` object
        An `mpl_toolkits.mplot3d.axes.Axes3D` object.
    P : ndarray of float
        `P` is an ndarray of PaletteViz coordinates (i.e. `|P| = n x 3`), 
    """
    # decide on what kind of axes to use
    if not ax:
        ax = Axes3D(plt.figure())

    euler = kwargs['euler'] if kwargs and 'euler' in kwargs else (-60, 30)
    title = kwargs['title'] if kwargs and 'title' in kwargs else None
    labels = kwargs['labels'] if kwargs and 'labels' in kwargs else None

    spread_factor = kwargs['spread_factor'] if kwargs and 'spread_factor' in kwargs else 'auto'
    inverted = kwargs['inverted'] if kwargs and 'inverted' in kwargs else True
    normalized = kwargs['normalized'] if kwargs and 'normalized' in kwargs else True
    project_collapse = kwargs['project_collapse'] if kwargs and 'project_collapse' in kwargs else True

    colorbar = kwargs['colorbar'] if kwargs and 'colorbar' in kwargs else None
    lims = kwargs['lims'] if kwargs and 'lims' in kwargs else (None, None, None)
    hide_layers = kwargs['hide_layers'] if kwargs and 'hide_layers' in kwargs else None

    anchor_linewidth = kwargs['anchor_linewidth'] if kwargs and 'anchor_linewidth' in kwargs else 1.0
    anchor_label_prefix = kwargs['anchor_label_prefix'] \
            if kwargs and 'anchor_label_prefix' in kwargs else r"$f_{{{:d}}}$"
    anchor_label_fontsize = kwargs['anchor_label_fontsize'] \
            if kwargs and 'anchor_label_fontsize' in kwargs else 'large'
    anchor_label_fontname = kwargs['anchor_label_fontname'] \
            if kwargs and 'anchor_label_fontname' in kwargs else None
    anchor_label_fontstyle = kwargs['anchor_label_fontstyle'] \
            if kwargs and 'anchor_label_fontstyle' in kwargs else 'normal'

    verbose = kwargs['verbose'] if kwargs and 'verbose' in kwargs else False

    # remove once they are read
    kwargs = pop(kwargs, 'euler')
    kwargs = pop(kwargs, 'title')
    kwargs = pop(kwargs, 'labels')
    kwargs = pop(kwargs, 'spread_factor')
    kwargs = pop(kwargs, 'inverted')
    kwargs = pop(kwargs, 'normalized')
    kwargs = pop(kwargs, 'project_collapse')
    kwargs = pop(kwargs, 'colorbar')
    kwargs = pop(kwargs, 'lims')
    kwargs = pop(kwargs, 'hide_layers')
    kwargs = pop(kwargs, 'anchor_linewidth')
    kwargs = pop(kwargs, 'anchor_label_prefix')
    kwargs = pop(kwargs, 'anchor_label_fontsize')
    kwargs = pop(kwargs, 'anchor_label_fontname')
    kwargs = pop(kwargs, 'anchor_label_fontstyle')
    kwargs = pop(kwargs, 'verbose')

    # decide on what kind of axes to use
    if ax is None:
        ax = Axes3D(plt.figure())
 
    # paletteviz coordinates
    if ax:
        if mode == 'star':
            if verbose:
                print("Plotting palette-star-viz.")
            P, K, _, Z = get_palette_star_coordinates(A, depth_contours=depth_contours, \
                                                        n_partitions=n_partitions, \
                                                        inverted=inverted, \
                                                        normalized=normalized, \
                                                        project_collapse=project_collapse, \
                                                        verbose=verbose)
        elif mode == 'radviz':
            if verbose:
                print("Plotting palette-radviz.")
            P, K, _, Z = get_palette_radviz_coordinates(A, depth_contours=depth_contours, \
                                                        n_partitions=n_partitions, \
                                                        spread_factor=spread_factor, \
                                                        normalized=normalized, \
                                                        project_collapse=project_collapse, \
                                                        verbose=verbose)
        else:
            raise ValueError("Unknown mode, it has to be one of {'star', 'radviz'}.")

        # if there is any layer to hide
        P_ = None # we keep a copy of the original data since they are going to change here.
        if hide_layers is not None:
            I = np.ones(P.shape[0]).astype(bool)
            for l in hide_layers:
                I[np.argwhere((P[:,-1] > Z[l] - 0.01) & (P[:,-1] < Z[l] + 0.01))] = False
            P_ = np.array(P, copy=True) # copy
            P = P[I]
            if c is not None:
                c = c[I]
            if s is not None:
                s = s[I]
            if labels is not None:
                labels = labels[I]
            Z = Z[np.array(list(set(range(Z.shape[0])) - set(hide_layers)))]

        # do the plot
        if labels is not None:
            if isinstance(labels, str):
                ax.scatter(P[:,0], P[:,1], P[:,2], s=s, c=c, label=labels, **kwargs)
            else:
                if isinstance(labels, np.ndarray): 
                    labels = labels.tolist()
                label_groups = group_labels_by_appearance(labels)
                for i,v in enumerate(label_groups):
                    ax.scatter(P[v[0],0], P[v[0],1], P[v[0],2], s=s[v[0]], c=c[v[0]], label=v[1], \
                            zorder=label_groups.shape[0]-i, **kwargs)
        else:
            ax.scatter(P[:,0], P[:,1], P[:,2], s=s, c=c, **kwargs)

        # set lims
        ax.set_xlim(lims[0])
        ax.set_ylim(lims[1])
        ax.set_zlim(lims[2])
        
        # set camera angle
        ax.view_init(euler[1], euler[0])

        # draw axes?
        if draw_axes:
            ax.set_axis_on()
        else:
            ax.set_axis_off()

        # draw anchors?
        if draw_anchors:
            polygon_layers = draw_anchors['polygons']
            label_layers = draw_anchors['labels']
            circle_layers = draw_anchors['circles']
            for i,z in enumerate(Z):
                if i in polygon_layers:
                    set_polar_anchors(ax, K, z=z, anchor_linewidth=anchor_linewidth)
                if i in label_layers:
                    if mode == 'radviz':
                        set_polar_anchor_labels(ax, K, z=z, \
                                        label_prefix=anchor_label_prefix, \
                                        label_fontsize=anchor_label_fontsize, \
                                        label_fontname=anchor_label_fontname, \
                                        label_fontstyle=anchor_label_fontstyle)
                    elif mode == 'star':
                        draw_circle = True if i in circle_layers else False
                        set_polar_anchor_labels(ax, K, z=z, draw_circle=draw_circle, \
                                        label_prefix=anchor_label_prefix, \
                                        label_fontsize=anchor_label_fontsize, \
                                        label_fontname=anchor_label_fontname, \
                                        label_fontstyle=anchor_label_fontstyle)
                    else:
                        raise ValueError("Unknown mode, it has to be one of {'star', 'radviz'}.")

        # colorbar?
        if colorbar and isinstance(colorbar, tuple) and len(colorbar) >= 2 \
                and isinstance(colorbar[0], np.ndarray) and isinstance(colorbar[1], np.ndarray):
            vmin,vmax = 0.0, 1.0
            cbc, cbg = colorbar[0], colorbar[1]
            cbl = colorbar[2] if len(colorbar) > 2 and colorbar[2] else None
            cbp = colorbar[3] if len(colorbar) > 3 and colorbar[3] else -0.05
            Id = np.column_stack((cbg,cbc)).astype(object)
            Id = Id[np.argsort(Id[:, 0])] 
            c, g = Id[:,1:].astype(float), Id[:,0].astype(float)
            vmin, vmax = np.min(g), np.max(g)
            norm = mc.Normalize(vmin=vmin, vmax=vmax)
            cmap = ListedColormap(c)
            if cbl:
                ax.figure.colorbar(cm.ScalarMappable(norm=norm, cmap=cmap), \
                        orientation='vertical', label=cbl, pad=cbp, shrink=0.5)
            else:
                ax.figure.colorbar(cm.ScalarMappable(norm=norm, cmap=cmap), \
                            orientation='vertical', pad=cbp, shrink=0.5) 

        # where to put the legend
        if labels:
            ax.legend(loc="best", ncol=2)
        
        # title?
        ax.set_title(title, pad=0.0)

        if P_ is not None:
            return (ax, P_)
        else:
            return (ax, P)
    else:
        raise TypeError("A valid `mpl_toolkits.mplot3d.axes.Axes3D` object is not found.")



"""star.py -- A customized and more flexible Star-coordinate plotting module. 

    This module provides a customized and more flexible function for Star-coordinate [1]_ 
    plotting. This module also provides different relevant fucntions, parameters and tools.

    Copyright (C) 2016
    Computational Optimization and Innovation (COIN) Laboratory
    Department of Computer Science and Engineering
    Michigan State University
    428 S. Shaw Lane, Engineering Building
    East Lansing, MI 48824-1226, USA
    
    References
    ----------
    .. [1] Eser Kandogan. 2001. Visualizing Multi-dimensional Clusters, Trends, 
        and Outliers Using Star Coordinates. In Proceedings of the Seventh ACM 
        SIGKDD International Conference on Knowledge Discovery and Data Mining 
        (KDD '01). ACM, New York, NY, USA, 107--116.

.. moduleauthor:: AKM Khaled Talukder <talukde1@msu.edu>

"""

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import matplotlib.colors as mc
from matplotlib.colors import ListedColormap
from viz.utils import transform as tr
from viz.plotting.utils import set_polar_anchors, set_polar_anchor_labels, pop, group_labels_by_appearance


__all__ = ["get_star_coordinates", "plot"]


def get_star_coordinates(X, inverted=True, normalized=True):
    r"""Generate Star-coordinates from data points `X`.

    Maps all the data points in `X` (i.e. `|X| = n x m`) onto 
    star-coordinate [1]_ positions. 

    Parameters
    ----------
    X : ndarray 
        `n` number of `m` dimensiomal points as input.
    inverted : bool, optional
        In the original star-coordinate plot, a point that is exactly on 
        an anchor means the point has the lowest value in that dimension.
        However, in our case, we need to make it reverted since we want 
        to place a point farther from an anchor point if the point has the
        lowest value in that dimension. Default 'True' when optional.
    normalized : bool, optional
        We need to normalize within `[0.0, 1.0]`. 
        However in any case, if someone is not opted.
        Default 'True' when optional.

    Returns
    -------
    P : ndarray of float
        An ndarray of PaletteViz coordinates (i.e. `|P| = n x 3`).
    K : ndarray of float
        The positions of the anchor points (i.e. `|K| = m x 2`),
    B : ndarray of float
        The lower bound and upper bound of `P`. 

    References
    ----------
    .. [1] Eser Kandogan. 2001. Visualizing Multi-dimensional Clusters, Trends, 
        and Outliers Using Star Coordinates. In Proceedings of the Seventh ACM 
        SIGKDD International Conference on Knowledge Discovery and Data Mining 
        (KDD '01). ACM, New York, NY, USA, 107--116.
    """

    m = X.shape[1]
    # check if the data needs to be normalized
    if normalized:
        X_ = tr.normalize(X, lb=np.zeros(m), ub=np.ones(m))
    else:
        X_ = X

    if inverted:
        X_ = 1.0 - X_
    
    T = 2 * np.pi * (np.arange(0, m, 1).astype(int) / m)
    COS,SIN = np.cos(T), np.sin(T)
    LB,UB = np.min(X_, axis=0), np.max(X_, axis=0)
    D = (UB - LB)
    U, V = (COS / D), (SIN / D)
    P = np.column_stack((np.sum(X_ * U, axis=1), np.sum(X_ * V, axis=1)))
    
    B = [np.min(P, axis=0), np.max(P, axis=0)]
    K = np.column_stack((COS, SIN))
    return (P, K, B)


def plot(A, ax=None, s=1, c=mc.TABLEAU_COLORS['tab:blue'], inverted=True, \
        normalized=True, draw_axes=False, draw_anchors=True, **kwargs):
    r"""A customized star-coordinate plot.

    This star-coordinate [1]_ plot is customized for the experiments. 
    
    Parameters
    ----------
    A : ndarray 
        `n` number of `m` dim. points to be plotted.
    ax : An `matplotlib.axes.Axes` object, optional
        Default `None` when optional.
    s : int or 1-D array_like, optional
        Point size, or an array of point sizes. Default 1 when optional.
    c : A `matplotlib.colors` object, str or an array RGBA color values.
        Colors to be used. Default `mc.TABLEAU_COLORS['tab:blue']` when 
        optional.
    inverted : bool, optional
        See `get_star_coordinates()` function for details.
    normalized : bool, optional
        If needed, the data points in `A` can be normalized within `[0.0, 1.0]`. 
        Default `True` when optional.
    draw_axes : bool, optional
        If `True`, the radviz plot will show axes. Default `False` when optional.
    draw_anchors : bool, optional
        If `False`, the radviz plot will hide anchors. Default `True` when optional.

    Other Parameters
    ----------------
    title : str, optional
        The plot title. Default `None` when optional.
    labels : str, array_like or list of str, optional
        A string or an array/list of strings for labeling each point. Which basically
        means the class label of each row. Default `None` when optional. This will be
        used to set the legend in the figure. If `None` there will be no legend.
    colorbar : (Cbc, Cbg, Cbl, Cbp) a tuple of (ndarray, ndarray, str, float), optional
        If a user wants to put a colorbar, a tuple `(Cbc, Cbg, Cbl)` tuple can be 
        provided. `Cbc` is an array of RGBA color values or an `matplotlib.colors` 
        object. The gradient of the colorbar is specified in `Cbg` which is an 1-D 
        array of float. Cbl is the label of the colorbar, a string. `Cbp` is the 
        colorbar padding width in `float`. `colorbar` is default `None` when optional. 
    label_prefix : str, optional
        See `set_anchor_labels()` function for details.
    label_fontsize : str or int, optional
        See `set_anchor_labels()` function for details.
    label_fontname : str, optional
        See `set_anchor_labels()` function for details.
    label_fontstyle : str, optional
        See `set_anchor_labels()` function for details.
    **kwargs : dict
        All other keyword args for matplotlib `scatter()` function.

    Returns
    -------
    ax : `matplotlib.axes.Axes` object
        A `matplotlib.axes.Axes` object.
    P : ndarray of float
        `P` is an ndarray of radviz coordinates (i.e. `|P| = n x 2`), 

    References
    ----------
    .. [1] Eser Kandogan. 2001. Visualizing Multi-dimensional Clusters, Trends, 
        and Outliers Using Star Coordinates. In Proceedings of the Seventh ACM 
        SIGKDD International Conference on Knowledge Discovery and Data Mining 
        (KDD '01). ACM, New York, NY, USA, 107--116.
    """

    # decide on what kind of axes to use
    if not ax:
        ax = fig.gca(plt.figure())
    
    title = kwargs['title'] if kwargs and 'title' in kwargs else None
    labels = kwargs['labels'] if kwargs and 'labels' in kwargs else None
    colorbar = kwargs['colorbar'] if kwargs and 'colorbar' in kwargs else None

    label_prefix = kwargs['label_prefix'] if kwargs and 'label_prefix' in kwargs else r"$f_{{{:d}}}$"
    label_fontsize = kwargs['label_fontsize'] if kwargs and 'label_fontsize' in kwargs else 'large'
    label_fontname = kwargs['label_fontname'] if kwargs and 'label_fontname' in kwargs else None
    label_fontstyle = kwargs['label_fontstyle'] if kwargs and 'label_fontstyle' in kwargs else 'normal'

    # remove once they are read
    kwargs = pop(kwargs, 'title')
    kwargs = pop(kwargs, 'labels')
    kwargs = pop(kwargs, 'colorbar')
    kwargs = pop(kwargs, 'label_prefix')
    kwargs = pop(kwargs, 'label_fontsize')
    kwargs = pop(kwargs, 'label_fontname')
    kwargs = pop(kwargs, 'label_fontstyle')

    if ax:
        P, K, [lb, ub] = get_star_coordinates(A, inverted=inverted, normalized=normalized)
        
        # do the plot
        if labels is not None:
            if isinstance(labels, str):
                ax.scatter(P[:,0], P[:,1], s=s, c=c, label=labels, **kwargs)
            else:
                if isinstance(labels, np.ndarray): 
                    labels = labels.tolist()
                label_groups = group_labels_by_appearance(labels)
                for i,v in enumerate(label_groups):
                    ax.scatter(P[v[0],0], P[v[0],1], s=s[v[0]], c=c[v[0]], label=v[1], \
                            zorder=label_groups.shape[0]-i, **kwargs)
        else:
            ax.scatter(P[:,0], P[:,1], s=s, c=c, **kwargs)        

        if draw_axes:
            ax.set_axis_on()
        else:
            ax.set_axis_off()
        
        if draw_anchors:
            set_polar_anchors(ax, K)
            set_polar_anchor_labels(ax, K, draw_circle=True, \
                        label_prefix=label_prefix, label_fontsize=label_fontsize, \
                        label_fontname=label_fontname, label_fontstyle=label_fontstyle)
        ax.set_xlim(lb[0] - 0.1 if lb[0] < -1 else -1.1, ub[0] + 0.1 if ub[0] > 1 else 1.1)
        ax.set_ylim(lb[1] - 0.1 if lb[1] < -1 else -1.1, ub[1] + 0.1 if ub[1] > 1 else 1.1)
        ax.set_aspect('equal')
        
        # colorbar?
        if colorbar and isinstance(colorbar, tuple) and len(colorbar) >= 2 \
                and isinstance(colorbar[0], np.ndarray) and isinstance(colorbar[1], np.ndarray):
            vmin,vmax = 0.0, 1.0
            cbc, cbg = colorbar[0], colorbar[1]
            cbl = colorbar[2] if len(colorbar) > 2 and colorbar[2] else None
            cbp = colorbar[3] if len(colorbar) > 3 and colorbar[3] else 0.05
            Id = np.column_stack((cbg,cbc)).astype(object)
            Id = Id[np.argsort(Id[:, 0])] 
            c, g = Id[:,1:].astype(float), Id[:,0].astype(float)
            vmin, vmax = np.min(g), np.max(g)
            norm = mc.Normalize(vmin=vmin, vmax=vmax)
            cmap = ListedColormap(c)
            if cbl:
                ax.figure.colorbar(cm.ScalarMappable(norm=norm, cmap=cmap), \
                        orientation='vertical', label=cbl, pad=cbp, shrink=0.5)
            else:
                ax.figure.colorbar(cm.ScalarMappable(norm=norm, cmap=cmap), \
                            orientation='vertical', pad=cbp, shrink=0.5) 

        # where to put the legend
        if labels is not None:
            ax.legend(loc="best", ncol=1)
        
        # title?
        ax.set_title(title, y=ax.get_ylim()[-1]-0.05)
        
        return (ax, P)
    else:
        raise TypeError("A valid `matplotlib.axes.Axes` object is not found.")



"""simple_shape.py -- A simple convex-hull shape extraction algorithm

    This module provides different utility functions for a `simple
    convex-hull shape extraction` algorithm. In the case of simple
    convex-hull, we first collapse a dimension (drop a column) in 
    the data-set, then we project the data points on a unit simplex, 
    then we apply qhull algorithm to find the convex-hull boundary 
    points. Each depth contour found from the qhull will be considered
    as the `layers` for palette-viz representation of the data points.

    Copyright (C) 2016
    Computational Optimization and Innovation (COIN) Laboratory
    Department of Computer Science and Engineering
    Michigan State University
    428 S. Shaw Lane, Engineering Building
    East Lansing, MI 48824-1226, USA

.. moduleauthor:: AKM Khaled Talukder <talukde1@msu.edu>

"""

import numpy as np
from scipy.spatial import ConvexHull

__all__ = ["depth_contours"]

def collapse(F, d=0):
    r""" Function to collapse a dimension of the data points `F`.

    A very simple function that drops a column in the data points `F`.

    Parameters
    ----------
    F : ndarray
        A sample of `n` data points in `m` dimension, i.e. `|F| = n x m`.
    d : int, optional
        The dimension to be dropped. Default 0 (i.e. first coordinate) 
        when optional.

    Returns
    -------
    F : ndarray
        Data points with one less column than that of `F`, i.e. `|F| = n x (m-1)` 

    """
    return F[:,np.delete(np.arange(0,F.shape[1]).astype(np.int64), d)]

def project(F):
    r""" Function to project data points `F` onto a sunit simplex.

    A very simple function that projects data points `F` onto a 
    unit simplex, where `|F| = n x m`.

    Parameters
    ----------
    F : ndarray
        A sample of `n` data points in `m` dimension, i.e. `|F| = n x m`.

    Returns
    -------
    P : ndarray
        Projected data points from `F`, i.e. `|P| = n x m`.

    """
    n,m = F.shape
    u = 1 / np.sqrt(np.ones(m) * m)
    uTF = np.sum(u.T * F, axis=1)
    uTuTFT = (np.tile(u, (n,1)).T * uTF).T
    P = (F - uTuTFT) + (u / np.sqrt(m))
    return P

def depth_contours(F, project_collapse=True, verbose=False):
    r"""Function to find all the depth-contours of the data point in `F`.

    This function applies convex-hull (i.e. qhull, `scipy.spatial.ConvexHull`)
    algorithm to find the first depth-contour of `F`. Then removes them and flag
    them as the layer 1. Then applies qhull on the remaining data points and keep
    doing the same until more than `2 * m + 1` data points are left.

    Parameters
    ----------
    F : ndarray
        A sample of `n` data points in `m` dimension, i.e. `|F| = n x m`.
    project_collapse : bool, optional
        In this shape finding algorithm, if `m > 3`, we first project the data points
        onto a unit simplex, then we collapse one dimension. Thus a 3D points become
        a 2D plane. If we set it to `False`, then this transformation will not occur.
        However, in that case, there might be just one layer if the points are on a
        full convex surface. Also the processing will be extremely slow if there are
        many high-dimensional (`m > 4`) data points. Therefore, setting it to `False`
        is not recommended. Default `True` when optional.
    verbose : bool, optional
        Verbose level. Default `False` when optional.

    Returns
    -------
    L : A jagged ndarray
        Projected data points from `F`, i.e. `|P| = n x m`.
    """
    n,m = F.shape
    if project_collapse:
        if verbose:
            print("Projecting on a simplex and collapsing.")
        P = project(F)
        if P.shape[1] >= 2:
            P = collapse(P, d=m-1)
    else:
        P = np.array(F, copy=True)

    Id = np.arange(0, n, 1).astype(int)
    # print("Id.shape:", Id.shape)
    G = P[Id]

    L = []
    i = 0
    while Id.shape[0] >= (2 * m + 1):
        if verbose:
            print("Computing depth contour {:d} ...".format(i))
        H = ConvexHull(G, qhull_options="Qa QJ Q12")
        Ih = H.vertices
        # print("i:", i, "Ih:", Ih, "Ih.shape:", Ih.shape)
        # print("i:", i, "Id[Ih]:", Id[Ih], "Id[Ih].shape:", Id[Ih].shape)
        L.append(Id[Ih])
        Id = np.delete(Id, Ih)
        if Id.shape[0] > 0:
            # print("i:", i, "Id.shape:", Id.shape, "\n")
            G = P[Id]
            i = i + 1
        else:
            break
    if Id.shape[0] > 0:
        L.append(Id)
    if verbose:
        print("Done.")

    return np.array(L, dtype=object)



import unittest
import numpy as np
from numpy import testing as npt
from viz.utils import dm

class TestDm(unittest.TestCase):
    A = np.diagflat([[1,2,3], [4,5,6]]).astype(int)
    B = np.arange(0,6,1).astype(int)
    x, y = np.mgrid[0:5, 5:10]
    C = np.c_[x.ravel(), y.ravel()]

    def test_nadir(self):
        y = dm.nadir(self.A)
        npt.assert_equal(y, np.array([1, 2, 3, 4, 5, 6]))
    
    def test_ideal(self):
        y = dm.ideal(self.A)
        npt.assert_equal(y, np.array([0, 0, 0, 0, 0, 0]))

    def test_knees(self):
        y = dm.knees(self.B)
        npt.assert_equal(y, np.array([5]))

    def test_tradeoff(self):
        y = dm.tradeoff(self.C)
        npt.assert_equal(y[0], np.array([np.nan,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,
        0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.]))
        npt.assert_equal(y[1], np.array([ 0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15, 16,
       17, 18, 19, 20, 21, 22, 23]))

if __name__ == '__main__':
    unittest.main()



"""utils.py -- A collection of different utility functions for plotting.
    
    This module provides different utility functions for plotting. For example,
    changing/setting point and line colors, point and line size, setting up
    2D/3D patches etc.

    Copyright (C) 2016
    Computational Optimization and Innovation (COIN) Laboratory
    Department of Computer Science and Engineering
    Michigan State University
    428 S. Shaw Lane, Engineering Building
    East Lansing, MI 48824-1226, USA

.. moduleauthor:: AKM Khaled Talukder <talukde1@msu.edu>

"""

import numpy as np
import matplotlib.cm as cm
import matplotlib.colors as mc
from matplotlib.patches import Circle
from mpl_toolkits.mplot3d import art3d
from matplotlib.patches import FancyArrowPatch
from mpl_toolkits.mplot3d import proj3d
from viz.utils import transform as tr

__all__ = ["resize_by_tradeoff", "default_color", "color_by_cv", \
            "color_by_dist", "enhance_color", \
            "set_polar_anchors", "set_polar_anchor_labels", \
            "pop", "Arrow3D", "group_labels_by_appearance", "cfs"]


"""
    Different color factors for different problems with constraint functions.
    Users can use these for better viewing. These were found empirically.
"""
cfs = {'dtlz8': { 2: 2.0, 3: 6.0, 4: 8.0, 6: 14.0, 8: 16.0}}


def group_labels_by_appearance(labels):
    r""" A function to partition the data according to the labels.

    Let's say we have `n` data points, `F = {p1, p2, p3, ..., pn}`.
    Also each data point has a label, and there are 2 kind of labels, 
    i.e. `L = {l1, l2, l1, l1, ..., l2}` where each label `li` corresponds
    to point `pi`. This function partitions `F` with respect to `L` and
    returns a list of lists `ul`, where `ul = [[li, [i1, i2, i3, ..., ip]], 
    [lj, [j1, j2, j3, ..., jq]] ... ]`. This means all the data points 
    indexed by `ii` have labels `li` and all the data points indexed by 'jj' 
    have labels `lj` and so on.
    
    Parameters
    ----------
    labels : list of str
        A list of strings specifying label for each data point.

    Returns
    -------
    ul : ndarray of object
        A jagged array of arrays. Where each array has two items.
        The first item is the label and the next item is an array
        of int for indices of data points with that label.
    """
    ld = {}
    for i,label in enumerate(labels):
        if label in ld:
            ld[label].append(i)
        else:
            ld[label] = []
    ul = []
    for k in ld.keys():
        ul.append([ld[k], k])
    sorted(ul, key = lambda v: len(v[0]), reverse=True)
    return np.array(ul, dtype=object)


def pop(d, k):
    r"""A safe popping from a dict.
    """
    r = dict(d)
    if k in d:
        del r[k]
    return r

# we might need to draw arrows.
class Arrow3D(FancyArrowPatch):
    r"""A 3D arrow class.

    This class is derived from `matplotlib.patches.FancyArrowPath`.
    
    Usage: A 3D arrow from `(x1,y1,z1)` to `(x2,y2,z2)` will look like this.

    ```
    a = Arrow3D([x1, x2], [y1, y2], [z1, z2], mutation_scale = 10, lw = 1.0, \
                arrowstyle = "-|>", color = 'black')
    ax.add_artist(a)
    ```
    """
    def __init__(self, xs, ys, zs, *args, **kwargs):
        FancyArrowPatch.__init__(self, (0,0), (0,0), *args, **kwargs)
        self._verts3d = xs, ys, zs

    def draw(self, renderer):
        xs3d, ys3d, zs3d = self._verts3d
        xs, ys, zs = proj3d.proj_transform(xs3d, ys3d, zs3d, renderer.M)
        self.set_positions((xs[0],ys[0]),(xs[1],ys[1]))
        FancyArrowPatch.draw(self, renderer)


def resize_by_tradeoff(Mu, k=None, minsize=2.0, maxsize=10.0, kminsize=3.0, kmaxsize=5.0):
    r"""Resize the points w.r.t. tradeoff values.

    If we need to resize the points with respect to the tradeoff values, we can use this 
    function. This function assumes `Mu` values are within `[0.0,1.0]`. 

    Since point size of 0 is not visible, we normalize them within `[minsize, maxsize]`.
    In order to capture the relative tradeoffs, we then scale the sizes by the power of 2.
    If the user wants to emphasize a specific set of points so that they 'stand out' 
    from all the rest of the points, they can be indexed with `k`. Those points will be 
    pronounced by normalizing them within `[kminsize, kmaxsize]` and then scale them by 
    the power of 2.
    
    Parameters
    ----------
    Mu : 1-D array_like
        A 1D-array of tradeoff values, float. Also `0.0 <= Mu <= 1.0`.
    k : 1-D array_like, optional
        A 1D-array of integer indices to be used to specify which points
        will be increased in size by `kfactor`. Default `None` when optional.
    minsize : float, optional
        The minimum allowable size of each point before the exponential 
        scaling. Default 2.0 when optional.
    maxsize : float, optional
        The maximum allowable size of each point before the exponential 
        scaling. Default 10.0 when optional.
    kminsize : float, optional
        The minimum allowable size of the points indexed by `k`, before 
        the exponential scaling. Default 3.0 when optional.
    kmaxsize : float, optional
        The maximum allowable size of the points indexed by `k`, before 
        the exponential scaling. Default 5.0 when optional.
    """
    
    S = np.power(tr.normalize(Mu, lb=np.array([minsize]), ub=np.array([maxsize])), 2)
    if k is not None:
        S[k] = np.power(tr.normalize(Mu[k], lb=np.array([kminsize]), ub=np.array([kmaxsize])), 3)
    return S


def default_color(n, c=mc.TABLEAU_COLORS['tab:blue'], alpha=1.0):
    r"""Get an array of RGBA color values for default coloring.

    In any case, if we need to revert the point colorings to the 
    default matplotlib (`mc.TABLEAU_COLORS['tab:blue']`) coloring 
    scheme, we can use this function.

    Parameters
    ----------
    n : int
        The length of the output array containing RGBA color values.
    c : A `matplotlib.colors` object, str or an array RGBA color values.
        Colors to be used. Default `mc.TABLEAU_COLORS['tab:blue']` when 
        optional.
    alpha : float, optional
        Alpha transparency value. Default 0.5 when optional.
        
    Returns
    -------
    C : ndarray
        An array of RGBA color values.
    """
    C = np.array([mc.to_rgba(c, alpha) for _ in range(n)])
    return C


def color_by_cv(CV, factor=0.8, alpha=0.5):
    r"""Generate an array of color values from CV.

    Generate an array of RGBA color values from an array of 
    cumulative constraint violation values. This function
    uses the color gradient from `cm.cool`.

    Parameters
    ----------
    CV : ndarray
        An nd-array of cumulative contstraint violation values.
    factor : float, optional
        If `factor = 1.0`, then we will use the actual gradient
        from `cm.cool`. When this value is smaller than 1.0, then
        the gradient will be shifted left (i.e. some portion from 
        the highest end of the color gradient will be skipped). 
        Default 0.8 when optional.
    alpha : float, optional
        Alpha transparency value. Default 0.5 when optional.
        
    Returns
    -------
    `C` : ndarray
        An array of RGBA color values.
    """
    CV_ = tr.normalize(CV, lb=0.0, ub=1.0)
    C = np.array([mc.to_rgba(cm.cool(v * factor), alpha) for v in CV_])
    return C


def color_by_dist(X, P, alpha=0.5, factor=1.75):
    r"""Generate an array of RGBA color values w.r.t distance of 'X' from 'P'

    Generate an array of RGBA color values for the corresponding points
    in `X` with respect to the their distances from a single point `P`.
    We generally use `P` as the center of mass of `X`, but can be used
    in other contexts.

    Parameters
    ----------
    X : ndarray
        An set of `m`-dimensional data points, i.e. `|X| = n x m` 
    P : 1D-array
        A point `P`, a 1-D array. This can be the center of mass of `X`.
    alpha : float, optional
        Alpha transparency value. Default 0.5 when optional.
    factor : float, optional
        If `factor = 1.0`, then we will use the actual gradient
        from `cm.winter_r`. When this value is smaller than 1.0, then
        the gradient will be shifted right (i.e. some portion from 
        the lowest end of the color gradient will be skipped). 
        Default 1.75 when optional. A user might want to try with 
        different values for factor.
        
    Returns
    -------
    C : ndarray
        An array of RGBA color values.
    D : array_like
        An array of distance values.
    """
    D = tr.normalize(np.linalg.norm(P - X, axis=1), lb=0.1, ub=1.0)
    C = np.array([mc.to_rgba(cm.winter_r(v * factor), alpha) for v in D])
    return C, D


def enhance_color(C, k, alpha=1.0, c=mc.TABLEAU_COLORS['tab:red']):
    r"""Enhance the color of selected data points.

    Given an array of RGBA color values `C`, this function will enhance
    all the points indexed by `Ik` by recoloring them with TABLEAU red
    color. Assuming that the color of the other points won't overlap with
    the enhanced points.
    
    Parameters
    ----------
    C: ndarray
        An array of RGBA color values as input.
    Ik : array_like of int
        An array of integer indices.
    alpha : float, optional
        Alpha transparency value. Default 0.5 when optional.
    color : RGB color value, optional
        The color to be used to enhance the points. 
        Default `mc.TABLEAU_COLORS['tab:red']` when optional.
        
    Returns
    -------
    C_ : ndarray
        An array of RGBA color values.
    """
    C_ = np.array(C, copy=True)
    C_[k] = np.array([mc.to_rgba(c, alpha) for _ in range(C[k].shape[0])])
    return C_


def set_polar_anchors(ax, A, z=None, anchor_linewidth=1.0):
    r"""Function to draw anchor points for radviz (and other related plots).

    This function draws anchor points for radviz and other related plots
    like Star-coordinate plots.

    Parameters
    ----------
    ax : `matplotlib.axes.Axes` object
        This object must be passed.
    A : ndarray 
        `m x 2` size of sin and cos values for the anchor coordinates.
    z : float, optional
        The z-coordinate values of the corresponding acnhor in 3D.
        Default 'None' when optional.
    anchor_linewidth : float, optional
        The width for anchor lines. Default 1.0 when optional.
    """
    tgc = mc.TABLEAU_COLORS['tab:gray']
    for i in range(0, A.shape[0]-1):
        if z is None:
            # draw one polygon line
            ax.plot([A[i,0], A[i + 1,0]], [A[i,1], A[i + 1,1]], c=tgc, alpha=1.0, \
                        linewidth=anchor_linewidth, linestyle='dashdot')
            # draw a pair of polygon points
            ax.scatter(A[i,0], A[i,1], c=tgc, marker='o', s=20.0, alpha=1.0)
        else:
            # draw one polygon line
            ax.plot([A[i,0], A[i + 1,0]], [A[i,1], A[i + 1,1]], zs=z, c=tgc, alpha=1.0, \
                        linewidth=anchor_linewidth, linestyle='dashdot')
            # draw a pair of polygon points
            ax.scatter(A[i,0], A[i,1], zs=z, c=tgc, marker='o', s=20.0, alpha=1.0)
    if z is None:
        # last polygon line
        ax.plot([A[-1,0], A[0,0]], [A[-1,1], A[0,1]], c=tgc, alpha=1.0, \
                    linewidth=anchor_linewidth, linestyle='dashdot')
        # last pair of polygon points
        ax.scatter(A[-1,0], A[-1,1], c=tgc, marker='o', s=20.0, alpha=1.0)
    else:
        # last polygon line
        ax.plot([A[-1,0], A[0,0]], [A[-1,1], A[0,1]], zs=z, c=tgc, alpha=1.0, \
                    linewidth=anchor_linewidth, linestyle='dashdot')
        # last pair of polygon points
        ax.scatter(A[-1,0], A[-1,1], zs=z, c=tgc, marker='o', s=20.0, alpha=1.0)


def set_polar_anchor_labels(ax, A, z=None, draw_circle=False, label_prefix=r"$f_{:d}$", \
                            label_fontsize='large', label_fontname=None, label_fontstyle=None):
    r"""Function to put anchor labels for radviz and other related plots.

    Function to draw anchor point labels for radivz and other related plots
    like Star-coordinate plots.

    Parameters
    ----------
    ax : `matplotlib.axes.Axes` object
        This object must be passed.
    A : ndarray 
        `m x 2` size of sin and cos values for the anchor coordinates.
    z : float, optional
        The z-coordinate values of the corresponding acnhor in 3D.
        Default 'None' when optional.
    draw_circle : bool, optional
        Draws a circum-cricle around the anchor points. 
        Default `False` when optional.
    label_prefix : str, optional
        The axis-label-prefix to be used, default `r"$f_{:d}$"` when optional.
    label_fontsize : str or int, optional
        The fontsize for the axes labels. Default `'large'` when optional.
    label_fontname : str, optional
        The fontname for the axes labels. Default `None` when optional.
    label_fontstyle : str, optional
        The fontstyle for the axes labels. Default `None` when optional.
    """
    tgc = mc.TABLEAU_COLORS['tab:gray']
    # now put all the corner labels, like f1, f2, f3, ... etc.
    for xy, name in zip(A, [label_prefix.format(i+1) for i in range(A.shape[0])]):
        if z is None:
            if xy[0] < 0.0 and xy[1] < 0.0:
                ax.text(xy[0] - 0.025, xy[1] - 0.025, s=name, ha='right', va='top', \
                        fontname=label_fontname, fontsize=label_fontsize, fontstyle=label_fontstyle)
            elif xy[0] < 0.0 and xy[1] >= 0.0: 
                ax.text(xy[0] - 0.025, xy[1] + 0.025, s=name, ha='right', va='bottom', \
                        fontname=label_fontname, fontsize=label_fontsize, fontstyle=label_fontstyle)
            elif xy[0] >= 0.0 and xy[1] < 0.0:
                ax.text(xy[0] + 0.025, xy[1] - 0.025, s=name, ha='left', va='top', \
                        fontname=label_fontname, fontsize=label_fontsize, fontstyle=label_fontstyle)
            elif xy[0] >= 0.0 and xy[1] >= 0.0:
                ax.text(xy[0] + 0.025, xy[1] + 0.025, s=name, ha='left', va='bottom', \
                        fontname=label_fontname, fontsize=label_fontsize, fontstyle=label_fontstyle)
        else:
            if xy[0] < 0.0 and xy[1] < 0.0:
                ax.text(xy[0] - 0.025, xy[1] - 0.025, z=z, s=name, ha='right', va='bottom', \
                        fontname=label_fontname, fontsize=label_fontsize, fontstyle=label_fontstyle)
            elif xy[0] < 0.0 and xy[1] >= 0.0: 
                ax.text(xy[0] - 0.025, xy[1] + 0.025, z=z, s=name, ha='right', va='bottom', \
                        fontname=label_fontname, fontsize=label_fontsize, fontstyle=label_fontstyle)
            elif xy[0] >= 0.0 and xy[1] < 0.0:
                ax.text(xy[0] + 0.025, xy[1] - 0.025, z=z, s=name, ha='left', va='bottom', \
                        fontname=label_fontname, fontsize=label_fontsize, fontstyle=label_fontstyle)
            elif xy[0] >= 0.0 and xy[1] >= 0.0:
                ax.text(xy[0] + 0.025, xy[1] + 0.025, z=z, s=name, ha='left', va='bottom', \
                        fontname=label_fontname, fontsize=label_fontsize, fontstyle=label_fontstyle)
    if draw_circle:
        p = Circle((0, 0), 1, fill=False, linewidth=0.8, color=tgc)
        ax.add_patch(p)
        if z:
            art3d.pathpatch_2d_to_3d(p, z=z)



"""cdebmdk.py -- A Module to Generate Points on a 'CDEBMDK' Problem
    
    This module provides functions to generate points on an m-dimensional 
    constrained Knee problem [1]_. We call this Pareto-optimal front as
    'CDEBMDK'.

    Copyright (C) 2016
    Computational Optimization and Innovation (COIN) Laboratory
    Department of Computer Science and Engineering
    Michigan State University
    428 S. Shaw Lane, Engineering Building
    East Lansing, MI 48824-1226, USA


    References
    ----------
    .. [1] Branke J., Deb K., Dierolf H., Osswald M. (2004) Finding Knees in 
        Multi-objective Optimization. In: Yao X. et al. (eds) Parallel Problem 
        Solving from Nature - PPSN VIII. PPSN 2004. Lecture Notes in Computer 
        Science, vol 3242. Springer, Berlin, Heidelberg.

.. moduleauthor:: AKM Khaled Talukder <talukde1@msu.edu>

"""

import numpy as np
from viz.generators import debmdk
from viz.utils import transform as tr

__all__ = ["surface"]

def cvf(F):
    r"""The constraint function.

    This procedure computes the constraint violation values on
    an objective vector 'F'.
    """
    epsilon = {2: 0.50, 3: 1.5, 4: 1.5, 8: 1.75}
    m = F.shape[0]
    return (np.sum([((f - 1.75) ** 2) for f in F[0:-1]]) + ((F[-1] - 2.1) ** 2)) \
            - (epsilon[m] * epsilon[m])

def get_feasible(F, X):
    r"""Filters out the feasible solutions from given data points 'F' (and 'X').

    This function computes the constraint violation values for each data points 
    in matrix 'F' and returns only the feasible points along with the corresponding
    design variable values in 'X'.
    """

    G = np.apply_along_axis(cvf, 1, F)
    If = np.where(G <= 0.0)[0]
    return F[If], X[If], G[If]

def surface(r=1, k=1, n=10, m=2, mode='lhc', **kwargs):
    r"""Generate `n` number of points on an 'CDEBMDK' problem [1]_.

    This problem is adapted from 'DEBMDK' problem, it applies a constraint function
    on 'DEBMDK' hyper-surface and samples points from it. The constraint function
    bounds the hypersurface to a smaller region. 

    The radius of the `m`-sphere is specified as 1. The total number of "knees"
    is specified in `k`. The point generation is currently done in two ways of 
    random sampling -- Latin Hypercube (LHC) and LHC with normalization. Other 
    ways will be added later. This function returns only the feasible solutions
    where the constraint function is specified in 'cvf()'.

    Parameters
    ----------
    k : int, optional
        The total number of knees. Default 1 when optional.
    n : int, optional
        The total number of points. Default 10 when optional.
    m : int, optional
        The dimension of the sphere. Default 2 when optional.
    mode : str, {'lhc', 'lhcl2', 'dd'}, optional
        If `mode = `lhc``, then LHC sampling will be used and points will be generated
        using standard spherical coordinate systems. If `mode = `lhcl2``, then we will
        use a normalized LHC sampling to generate uniformly distributed points on the
        sphere using the method described in [2]_. If 'mode = 'dd', then we will generate
        points using the subproblem decomposition technique used in NBI method 
        (a.k.a. "Das-Dennis's Approach") discussed in [3]_. Default first when optional.

    Other Parameters
    ----------------
    delta : float, optional
        `delta` value for normalized LHC, this is used so that we only keep vectors
        `V` such that `np.linalg.norm(V, 1) > delta`. The default value is 0.0001 but
        you might want to change it according to your application.

    Returns
    -------
    F : ndarray
        `n` points on the `m`-sphere, i.e. `|F| = n x m`.
    X : ndarray
        `n` points of the `m-1` dimensional spherical coordinate values, i.e. `|X| = n x (m-1)`.

    References
    ----------
    .. [1] Branke J., Deb K., Dierolf H., Osswald M. (2004) Finding Knees in 
        Multi-objective Optimization. In: Yao X. et al. (eds) Parallel Problem 
        Solving from Nature - PPSN VIII. PPSN 2004. Lecture Notes in Computer 
        Science, vol 3242. Springer, Berlin, Heidelberg.

    .. [2] Simon C., "Generating uniformly distributed numbers on a sphere," online: 
        http://corysimon.github.io/articles/uniformdistn-on-sphere/ 

    .. [3] I. Das and J. E. Dennis, "Normal-Boundary Intersection: A New Method for 
        Generating the Pareto Surface in Nonlinear Multicriteria Optimization Problems," 
        SIAM Journal on Optimization, vol. 8, (3), pp. 631-27, 1998.
    """
    # This is needed for debmdk and dtlz2
    delta = kwargs['delta'] if (len(kwargs) > 0 and 'delta' in kwargs) else 0.0001
    
    F, X = debmdk.surface(r=r, n=n, m=m, mode=mode, delta=delta)
    F, X, G = get_feasible(F, X)
    CV = tr.normalize(G)
    return F, X, G, CV




"""io.py -- Different utility functions for I/O

    This module provides different utility functions for I/O.

    Copyright (C) 2016
    Computational Optimization and Innovation (COIN) Laboratory
    Department of Computer Science and Engineering
    Michigan State University
    428 S. Shaw Lane, Engineering Building
    East Lansing, MI 48824-1226, USA

.. moduleauthor:: AKM Khaled Talukder <talukde1@msu.edu>

"""

import os
import ast
import numpy as np

__all__ = ["loadtxt", "savetxt", "is_number"]

def is_num(s):
    r""" Inner function for `is_number()`
    """
    return s[1:].replace('.','',1).isdigit() \
            if s[0] == '-' else s.replace('.','',1).isdigit()
        
def is_number(s):
    r""" Returns True is string is a number. 
    
    This is a better solution than all the approaches described here:
    https://stackoverflow.com/questions/354038/how-do-i-check-if-a-string-is-a-number-float  
    
    Parameters
    ----------
    s : str
        A string representing a number, like '-123', '12.3', '12e-06' etc.
    
    Returns
    -------
    bool : bool
    """
    if is_num(s):
        return True
    else:
        p = s.split('e')
        if len(p) != 2:
            return False
        else:
            return (is_num(p[0]) and is_num(p[1]))

def cast(x, dtype):
    r"""Typecasting a scalar value `x`.

    A very crude implementation of typecasting 
    a scalar `x` with respect to `dtype`.

    Parameters
    ----------
    x : str
        A scalar variable `x`
    dtype : data-type
        A data-type literal, can be `int`, `float`, `str` etc.

    Returns
    -------
    dtype(x) : typecasted `x`
    
    """
    if dtype is float:
        return float(x)
    elif dtype is int:
        return int(x)
    elif dtype is str:
        return str(x)
    else:
        # fall back
        return ast.literal_eval(x)

def tolist(A):
    r""" Recursively apply `tolist()` on an numpy array
    
    This fucnction recursively applies `numpy.tolist()` on an numpy array.
    It will work for any array, even on a jagged one.

    Parameters
    ----------
    A : An ndarray or a list or  alist of ndarrays or ndarrays of lists
        The input `ndarray` or `list` to be converted.

    Returns
    -------
    list : A list or list of lists
        Returns a fully converted list.
    """
    if isinstance(A, np.ndarray):
        return tolist(A.tolist())
    elif isinstance(A, list):
        return [tolist(a) for a in A]
    elif isinstance(A, tuple):
        return tuple(tolist(a) for a in A)
    else:
        return A

def loadtxt(fname, dtype=float, delimiter=None):
    r"""Load an array from a file, very similar to `npyio.loadtxt()`.

    Since `npyio.loadtxt()` is not useful when we want to load a jagged array
    from a text file, we have added this function exactly for that reason.    

    Parameters
    ----------
    fname : str or pathlib.Path
        A filename or a file path in string or `pathlib.Path` to read.
    dtype : data-type, optional
        Data-type of the resulting array. It can be `int`, `float`, `str` etc.
        See `cast()` function for details. Default `float` when optional. 
    delimiter : str, optional
        The string used to separate values. Default is whitespace when optional.

    Returns
    -------
    X : ndarray
        Data read from the text file.
    """
     
    if os.path.exists(fname):
        try:
            fp = open(fname, 'r')
            X = []
            for l in fp:
                a = [cast(v.strip(), dtype) for v in l.strip().split(delimiter)]
                X.append(np.array(a))
            fp.close()
            return np.array(X, dtype=object)
        except IOError:
            print("Coudn't open file {:s}".format(fname))
    else:
        raise OSError("File {0:s} not found.".format(fname))

def savetxt(fname, X, fmt='{:.18e}', delimiter=' ', newline='\n'):
    r"""Save an array to a file, very similar to `npyio.savetxt()`.

    Since `npyio.savetxt()` is not useful when we want to save a jagged array
    in a text format, we have added this function exactly for that reason.

    Parameters
    ----------
    fname : str
        A file path in string.
    X : 1D or 2D array_like
        Data to be saved to a text file. It can be a jagged array.
    fmt : str or sequence of str's, optional
        Python3 formatting string, i.e. '{:d}' for `int`, `{:f}` for `float` etc.
        Default '{:.18e}' when optional.
    delimiter : str, optional
        String or character separating columns. A whitespace when default.
    newline : str, optional
        String or character separating lines. Default `\n` when optional.
    """
    try:
        fp = open(fname, 'w+')
        for i in range(X.shape[0]-1):
            fp.write(delimiter.join([fmt.format(v) for v in X[i]]) + newline)
        fp.write(delimiter.join([fmt.format(v) for v in X[X.shape[0]-1]]))
        fp.close()
    except IOError:
        print("Coudn't open file {:s}".format(fname))



"""dtlz8.py -- A Module to Generate Points on a 'DTLZ8' Problem
    
    This module provides functions to generate points on a 'DTLZ8' problem [1]_.

    Copyright (C) 2016
    Computational Optimization and Innovation (COIN) Laboratory
    Department of Computer Science and Engineering
    Michigan State University
    428 S. Shaw Lane, Engineering Building
    East Lansing, MI 48824-1226, USA

    References
    ----------
    .. [1] K. Deb, L. Thiele, M. Laumanns and E. Zitzler. Scalable Multi-Objective 
        Optimization Test Problems. CEC 2002, p. 825 - 830, IEEE Press, 2002.

.. moduleauthor:: AKM Khaled Talukder <talukde1@msu.edu>

"""

import numpy as np
from viz.utils import sampling as smp
from viz.utils import transform as tr

__all__ = ["surface"]

# Suggested values for the parameters:
#    m, nl, ns = 2, 50, 750
#    m, nl, ns = 3, 100, 11000
#    m, nl, ns = 4, 400, 80000
#    m, nl, ns = 6, 600, 160000
#    m, nl, ns = 8, 800, 160000

def surface(m=2, nl=50, ns=750, mode='lhc', **kwargs):
    r"""Generate `nl + ns` number of points on 'DTLZ8' problem.
    
    This function generates points on the Pareto-optimal front of 'DTLZ8' 
    problem [1]_. The Pareto-optimal hypersurface is composed of a 3 dimensional 
    (2 dimensional in the case of 'm = 2') line and an 'm'-dimensional hypersurface.

    Since we are not solving the actual 'DTLZ8' problem, this function samples
    points directly on the Pareto-optimal hypersurface. As a result this function
    does not return the design variables.

    Parameters
    ----------
    m : int, optional
        The dimension of the Pareto-optimal front. Default 2 when optional.
    nl : int, optional
        The number of points on the 3 (or 2)-dimensional line. Default 50 when optional.
    ns : int, optional
        The number of points on the 'm'-dimensional hypersurface. Default 750 when optional.
    mode : str, {'lhc', 'lhcl2', 'grid'}, optional
        If `mode = `lhc``, then LHC sampling will be used and points will be generated
        using standard spherical coordinate systems. If `mode = `lhcl2``, then we will
        use a normalized LHC sampling to generate uniformly distributed points on the
        sphere using the method described in [2]_. If 'mode = 'grid', then we will 
        generate points using a grid. 'Default first when optional.

    Other Parameters
    ----------------
    delta : float, optional
        `delta` value for normalized LHC, this is used so that we only keep vectors
        `V` such that `np.linalg.norm(V, 1) > delta`. The default value is 0.0001 but
        you might want to change it according to your application.
    feasible_only : bool, optional
        When 'feasible_only = True', the function will return only the feasible solutions.
        Otherwise it will return all the points. Default 'True' when optional.

    Returns
    -------
    F : ndarray
        `n` number of points in the `m`-dimension, i.e. `|F| = n x m`.
    X : ndarray
        `n` number of points in `m-1`-dimension, i.e. `|X| = n x (m-1)`.
    G : ndarray
        'n' constraint violation values for each point. Where `|G| = n x m`.
        Since there are 'm' constraints. 'G[i,j] < 0' means infeasible and
        'G[i,j] >= 0' means feasible.
    CV : ndarray
        'n' cumulative constraint violation values. Where '|CV| = n x 1'. 'CV'
        is just the row wise sums of 'G'.

    References
    ----------
    .. [1] K. Deb, L. Thiele, M. Laumanns and E. Zitzler. Scalable Multi-Objective 
        Optimization Test Problems. CEC 2002, p. 825 - 830, IEEE Press, 2002.

    .. [2] Simon C., "Generating uniformly distributed numbers on a sphere," online: 
        http://corysimon.github.io/articles/uniformdistn-on-sphere/ 
    """
    
    mkeys = {'lhc', 'lhcl2', 'grid'}
    if mode not in mkeys:
        raise KeyError("'mode' has to be any of {:s}.".format(str(mkeys)))    

    feasible_only = kwargs['feasible_only'] if (len(kwargs) > 0 and 'feasible_only' in kwargs) else True
    delta = kwargs['delta'] if (len(kwargs) > 0 and 'delta' in kwargs) else 0.0001

    t = {2: 0, 3: 0, 4: -0.15, 6: -0.7, 8: -1.5}
    d = {2: m + 1, 3: m + 1, 4: m, 6: m - 2.5, 8: m - 4.5}

    pt1 = np.zeros((1, m))
    pt1[0,-1] = 1.0
    pt2 = np.ones((1, m)) / d[m]

    delta = 1 / nl

    L = (np.ones((nl, 1)) * pt1) \
        + (np.arange(0, 1, delta) * (pt2 - pt1).T).T
    # print("L.shape", L.shape)

    if mode == 'lhcl2':
        S = (np.ones((ns, m - 1)) / (m + 1)) \
            + (smp.lhcl2(ns, m - 1, delta = delta) * 1 - (1.0 / (m + 1)))
    elif mode == 'lhc':
        S = (np.ones((ns, m - 1)) / (m + 1)) \
            + (smp.lhc(ns, m - 1) * 1 - (1.0 / (m + 1)))
        print(smp.lhc(ns, m-1).shape)
    elif mode == 'grid':
        Sg = smp.grid(ns, m - 1)
        ns = Sg.shape[0]
        S = (np.ones((ns, m - 1)) / (m + 1)) \
            + (Sg * 1 - (1.0 / (m + 1)))

    # print("S.shape", S.shape)

    Fs = np.append(S, np.zeros((S.shape[0], 1)), 1)
    # print("Fs.shape", Fs.shape)
    Fs[:,-1] = (1 - np.sum(S, axis = 1)) / 2.0
 
    # for i in range(ns):
    #     Fs[i,-1] = (1 - np.sum(S[i])) / 2.0
    #     for j in range(m-1):
    #         Gs[i,j] = Fs[i,-1] - (1.0 - (m * Fs[i,j])) # infeasible < 0, feasible >= 0
    #     Gs[i,-1] = Fs[i,-1] - t[m] # infeasible < 0, feasible >= 0

    Gs = np.zeros(Fs.shape)
    # print("Gs.shape", Gs.shape)
    Gs[:,0:m-1] = (Fs[:,-1] - (1.0 - m * Fs[:,0:m-1]).T).T
    Gs[:,-1] = Fs[:,-1] - t[m]
    
    F = np.concatenate((L, Fs), axis=0)
    
    # For X, we assume the design variables are the
    # projection of F onto m-1 dimensions where the 
    # m-th dimension being collapsed.
    X = F[:,:-1]

    # print("F.shape", F.shape)
    
    maxG = np.max(Gs, axis = 0)
    # print("maxG", maxG)
    Gl = np.tile(maxG, (L.shape[0], 1)) 
    # print("Gl", Gl)
    
    G = np.concatenate((Gl, Gs), axis=0)
    # print("G.shape", G.shape)

    if feasible_only:
       If = np.where(~np.any(G < 0, axis=1))
       F = F[If]
       G = G[If]
       X = X[If]
    
    CV = tr.normalize(np.sum(G, axis=1))
    # print("CV.shape", CV.shape)
    # print("CV", CV)

    # We don't have X for this, since we are sampling the surface itself.
    # X = None

    return F, X, G, CV



"""c2dtlz2.py -- A Module to Generate Points on a C2-DTLZ2 Pareto-optimal front.
    
    This module provides functions to generate points on 'm'-dimensional Pareto-front
    on a C2-DTLZ2 problem [1]_. 

    Copyright (C) 2016
    Computational Optimization and Innovation (COIN) Laboratory
    Department of Computer Science and Engineering
    Michigan State University
    428 S. Shaw Lane, Engineering Building
    East Lansing, MI 48824-1226, USA

    References
    ----------
    .. [1] H. Jain and K. Deb, "An Evolutionary Many-Objective Optimization Algorithm Using 
        Reference-Point Based Nondominated Sorting Approach, Part II: Handling Constraints 
        and Extending to an Adaptive Approach," in IEEE Transactions on Evolutionary Computation, 
        vol. 18, no. 4, pp. 602-622, Aug. 2014, doi: 10.1109/TEVC.2013.2281534. 

.. moduleauthor:: AKM Khaled Talukder <talukde1@msu.edu>

"""

import numpy as np
from viz.generators import dtlz2
from viz.utils import transform as tr

__all__ = ["surface"]

def cvf(F):
    r"""The constraint function.

    This procedure computes the constraint violation values on
    an objective vector 'F'.
    """
    m = F.shape[0]
    if m == 2:
        r = 0.25
    elif m == 3:
        r = 0.4
    else:
        r = 0.5
    r1 = 0.50 * r
    lhs = min([((f - 1.0) ** 2) + np.sum([(F[j] ** 2) for j in range(m) if j != i]) - (r ** 2.0) \
            for i,f in enumerate(F)])
    rhs = np.sum([((f - (1.0 / (m ** 0.5))) ** 2) for f in F]) - (r1 ** 2)
    c = min(lhs, rhs)
    return c

def get_feasible(F, X):
    r"""Filters out the feasible solutions from given data points 'F' (and 'X').

    This function computes the constraint violation values for each data points 
    in matrix 'F' and returns only the feasible points along with the corresponding
    design variable values in 'X'.
    """

    G = np.apply_along_axis(cvf, 1, F)
    If = np.where(G <= 0.0)[0]
    return F[If], X[If], G[If]

def surface(r=1, n=10, m=2, mode='lhc', **kwargs):
    r"""Generate `n` number of points on split `m`-shpere.

    The details of 'C2-DTLZ2' is described in [1]_.

    The radius of the `m`-sphere is specified in `r`. The point generation is 
    currently done in two ways of random sampling -- Latin Hypercube (LHC) and
    LHC with normalization. Other ways will be added later.

    Parameters
    ----------
    r : float, optional
        The radius of the sphere. Default 1 when optional.
    n : int, optional
        The total number of points. Default 10 when optional.
    m : int, optional
        The dimension of the sphere. Default 2 when optional.
    mode : str, {'lhc', 'lhcl2', 'dd'}, optional
        If `mode = `lhc``, then LHC sampling will be used and points will be generated
        using standard spherical coordinate systems. If `mode = `lhcl2``, then we will
        use a normalized LHC sampling to generate uniformly distributed points on the
        sphere using the method described in [2]_. If 'mode = 'dd', then we will generate
        points using the subproblem decomposition technique used in NBI method 
        (a.k.a. "Das-Dennis's Approach") discussed in [3]_. Default first when optional.

    Other Parameters
    ----------------
    delta : float, optional
        `delta` value for normalized LHC, this is used so that we only keep vectors
        `V` such that `np.linalg.norm(V, 1) > delta`. The default value is 0.0001 but
        you might want to change it according to your application.

    Returns
    -------
    F : ndarray
        `n` points on the `m`-sphere, i.e. `|F| = n x m`.
    X : ndarray
        `n` points of the `m-1` dimensional spherical coordinate values, i.e. `|X| = n x (m-1)`.

    References
    ----------
    .. [1] H. Jain and K. Deb, "An Evolutionary Many-Objective Optimization Algorithm Using 
        Reference-Point Based Nondominated Sorting Approach, Part II: Handling Constraints 
        and Extending to an Adaptive Approach," in IEEE Transactions on Evolutionary Computation, 
        vol. 18, no. 4, pp. 602-622, Aug. 2014, doi: 10.1109/TEVC.2013.2281534.

    .. [2] Simon C., "Generating uniformly distributed numbers on a sphere," online: 
        http://corysimon.github.io/articles/uniformdistn-on-sphere/ 

    .. [3] I. Das and J. E. Dennis, "Normal-Boundary Intersection: A New Method for 
        Generating the Pareto Surface in Nonlinear Multicriteria Optimization Problems," 
        SIAM Journal on Optimization, vol. 8, (3), pp. 631-27, 1998.
    """
    # This is needed for dtlz2
    delta = kwargs['delta'] if (len(kwargs) > 0 and 'delta' in kwargs) else 0.0001

    F, X = dtlz2.surface(r=r, n=n, m=m, mode=mode, delta=delta) 
    F, X, G = get_feasible(F, X)
    CV = tr.normalize(G)
    return F, X, G, CV



import setuptools
from setuptools import setup
import os

_here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(_here, 'README.md'), encoding='utf-8') as f:
    print("long_description path = " + f.name)
    long_description = f.read()

version = {}
with open(os.path.join(_here, 'viz', 'version.py')) as f:
    print("version path = " + f.name)
    exec(f.read(), version)

pack = setuptools.find_packages()
print("packages =", pack)

setup(
    name='pviz',
    version=version['__version__'],
    author='AKM Khaled Talukder',
    author_email='talukde1@msu.edu',
    description=(\
            'A framework for high-dimensional (3 <= m <= 10) Pareto-optimal ' \
            + 'front visualization and analytics'),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/chudur-budur/pviz',
    license='Apache-2.0',
    # packages=['viz', 'viz.utils', 'viz.generators', 'viz.tda', 'viz.plotting'],
    packages=setuptools.find_packages(),
    setup_requires=['scipy==1.5.2', 'matplotlib'],
    install_requires=['scipy==1.5.2', 'matplotlib'],
    include_package_data=True,
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3.8'
        ],
    )



"""pcp.py -- A customized and more flexible Parallel-coordinate plotting module. 

    This module provides a customized and more flexible function for Parallel-coordinate 
    Plot (PCP) [1]_ visualization. This module also provides different relevant fucntions, 
    parameters and tools.

    Copyright (C) 2016
    Computational Optimization and Innovation (COIN) Laboratory
    Department of Computer Science and Engineering
    Michigan State University
    428 S. Shaw Lane, Engineering Building
    East Lansing, MI 48824-1226, USA
    
    References
    ----------
    .. [1] A. Inselberg and T. Avidan, "Classification and visualization for high-dimensional 
        data", Proc. 6th ACM SIGKDD Int. Conf. Knowledge Discovery and Data Mining (KDD ‘00), 
        pp. 370-374, 2000.

.. moduleauthor:: AKM Khaled Talukder <talukde1@msu.edu>

"""

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import matplotlib.colors as mc
from matplotlib.colors import ListedColormap
from viz.plotting.utils import pop 
from viz.utils import transform as tr
from viz.utils import dm

__all__ = ["plot"]

xmargins = {10: [0.7, 0.3], 9:[0.6, 0.3], 8:[0.5, 0.3], 7:[0.45, 0.3], 6:[0.3, 0.2]}
ymargins = {10: [-0.1, 0.08], 9:[-0.09, 0.075], 8:[-0.09, 0.075], 7:[-0.08, 0.075], 6:[-0.08, 0.075], 5:[-0.8, 0.08], 4:[-0.08, 0.08], 3:[-0.08, 0.08]}

def is_xticklabels_off(ax):
    r"""Checks if axes has already xtick labels

    Checks if an `matplotlib.axes.Axes` object has already
    xtick labels. 

    Parameters
    ----------
    ax : matplotlib axes object
        An `matplotlib.axes.Axes` object.
    
    Returns
    -------
    True/False : bool
    """

    xtl = ax.get_xticklabels()
    for s in xtl:
        if str(s) != "Text(0, 0, \'\')":
            return False
    return True

def get_yaxis_bounds(A):
    r"""
    """
    ub = dm.nadir(A)
    lb = dm.ideal(A)
    ubs = ["{:1.1e}".format(v) for v in ub]
    lbs = ["{:1.1e}".format(v) for v in lb]
    return [lbs, ubs]

def plot(A, ax=None, show_bounds=True, c=mc.TABLEAU_COLORS['tab:blue'], lw=1.0, labels=None, \
        xtick_labels=None, draw_vertical_lines=True, draw_grid=False, **kwargs):
    r"""A customized and more enhanced Parallel-coordinate plot.

    This Parallel-coordinate plot (PCP) [1]_ is customized for the experiments. 
    A lot of settings are customizable and configurable. Also it gives more 
    flexibility to the user compared to similar functions implemented in other 
    libraries like Pandas and seaborn. 
    
    Parameters
    ----------
    A : ndarray 
        `n` number of `m` dim. points to be plotted.
    ax : An `mpl_toolkits.mplot3d.axes.Axes3D` object, optional
        Default `None` when optional.
    show_bounds : bool, optional
        If `True` then the plot will show the lower and upper bounds of each data
        point (i.e. lines). Default `False` when optional.
    c : A `matplotlib.colors` object, str or an array RGBA color values.
        Colors to be used. Default `mc.TABLEAU_COLORS['tab:blue']` when 
        optional.
    lw : float, optional
        The line-width of each line in PCP. Default 1.0 when optional.
    labels : str, array_like or list of str, optional
        A string or an array/list of strings for labeling each line. Which basically
        means the class label of each row. Default `None` when optional. This will be
        used to set the legend in the figure. If `None` there will be no legend.
    xtick_labels : str, array_like or list of str, optional
        A string or an array/list of strings for xtick labels, for each column.
        Default `None` when optional. In that case, the labels will be `f_0`, `f_1` etc.
    draw_vertical_lines : bool, optional
        Decide whether we are going to put vertical y-axis lines in the plot for each
        column/feature. Default `True` when optional.
    draw_grid : bool, optional
        Decide whether we are going to put x-axis grid-lines in the plot. Default
        `False` when optional.

    Other Parameters
    ----------------
    title : str, optional
        The title of the figure. Default `None` when optional.
    column_indices : array_like or list of int, optional
        The indices of the columns of `A` to be plotted. Default `None` when optional.
    colorbar : (Cbc, Cbg, Cbl) a tuple of two ndarray and a str, optional
        If a user wants to put a colorbar, a tuple `(Cbc, Cbg, Cbl)` can be provided. 
        `Cbc` is an array of RGBA color values or an `matplotlib.colors` object. The 
        gradient of the colorbar is specified in `Cbg` which is an 1-D array of float. 
        Cbl is the label of the colorbar, a string. Default `None` when optional.
    axvline_width : float, optional
        The width of the vertical lines. Default 1.0 when optional.
    axvline_color : A `matplotlib.colors` object, str or an array RGBA color values.
        The color of the vertical lines. Default `black` when optional.
    **kwargs : dict
        All other keyword args for matplotlib `plot()` function.

    Returns
    -------
    ax : `mpl_toolkits.mplot3d.axes.Axes3D` object
        An `mpl_toolkits.mplot3d.axes.Axes3D` object.

    References
    ----------
    .. [1] A. Inselberg and T. Avidan, "Classification and visualization for high-dimensional 
        data", Proc. 6th ACM SIGKDD Int. Conf. Knowledge Discovery and Data Mining (KDD ‘00), 
        pp. 370-374, 2000.
    """
    
    # collect extra kwargs
    title = kwargs['title'] if kwargs and 'title' in kwargs else None    
    column_indices = kwargs['column_indices'] if kwargs and 'column_indices' in kwargs else None   
    colorbar = kwargs['colorbar'] if kwargs and 'colorbar' in kwargs else None
    axvline_width = kwargs['axvline_width'] if kwargs and 'axvline_width' in kwargs else 1.0
    axvline_color = kwargs['axvline_color'] if kwargs and 'axvline_color' in kwargs else 'black'
    
    # remove once they are read
    kwargs = pop(kwargs, 'title')
    kwargs = pop(kwargs, 'column_indices')
    kwargs = pop(kwargs, 'colorbar')
    kwargs = pop(kwargs, 'axvline_width')
    kwargs = pop(kwargs, 'axvline_color')
    
    if not ax:
        ax = plt.figure().gca()        

    lbs, ubs = get_yaxis_bounds(A)
    F = tr.normalize(A, lb=np.zeros(A.shape[1]), ub=np.ones(A.shape[1]))

    # build color list for each data point
    if (not isinstance(c, list)) and (not isinstance(c, np.ndarray)):
        c_ = c
        c = np.array([c_ for _ in range(F.shape[0])])
    elif (isinstance(c, list) and len(c) != F.shape[0]) \
        or (isinstance(c, np.ndarray) and c.shape[0] != F.shape[0]):
            raise ValueError("The length of c needs to be same as F.shape[0].")
        
    # build linewidth list for each data point
    if (not isinstance(lw, list)) and (not isinstance(lw, np.ndarray)):
        lw_ = lw
        lw = np.array([lw_ for _ in range(F.shape[0])])
    elif (isinstance(lw, list) and len(lw) != F.shape[0]) \
        or (isinstance(lw, np.ndarray) and lw.shape[0] != F.shape[0]):
            raise ValueError("The length of lw needs to be same as F.shape[0].")

    # get a list of column indices
    if column_indices:
        x = np.array(column_indices)
    else:
        x = np.arange(0,F.shape[1],1).astype(int)
    if len(x) < 2:
        raise ValueError("column_indices must be of length > 1.")
            
    # get a list of xtick_labels
    if xtick_labels is None:
        xtick_labels = ["$f_{:d}$".format(i) for i in range(F.shape[1])]
        
    # get a list of line labels, i.e. class labels
    if labels is not None and isinstance(labels, str):
        label = labels
        labels = np.array([label for _ in range(F.shape[0])])
        
    # draw the actual plot
    used_legends = set()
    for i in range(F.shape[0]):
        y = F[i,x]
        if labels is not None:
            label = labels[i]
            if label not in used_legends:
                used_legends.add(label)
                ax.plot(x, y, color=c[i], label=label, linewidth=lw[i], **kwargs)
            else:
                ax.plot(x, y, color=c[i], linewidth=lw[i], **kwargs)
        else:
            ax.plot(x, y, color=c[i], linewidth=lw[i], **kwargs)
    
    # decide on vertical axes
    if draw_vertical_lines:
        for i in x:
            ax.axvline(i, linewidth=axvline_width, color=axvline_color)

    # draw grid?
    if draw_grid:
        ax.grid()
    else:
        ax.spines['right'].set_visible(False)
        ax.spines['left'].set_visible(False)
 
    # decide on xtick_labels
    if xtick_labels is not None:
        if is_xticklabels_off(ax):
            ax.set_xticks(x)
            ax.set_xticklabels(xtick_labels)
            # Now completely change the axis ticks and labels
            # if there are bounds to be shown
            if show_bounds:
                ax.set_yticks([])
                ax.set_yticklabels([])
                plt.setp(ax.get_xticklabels(), fontsize=11, 
                        rotation=-45, ha="left", rotation_mode="anchor")
                ax.set_ylim([-0.1, 1.1])
                bottom, top = -0.1 + ymargins[F.shape[1]][0], 1.1 + ymargins[F.shape[1]][1]
                for i in range(A.shape[1]):
                    ax.text(i + ((0.68/10) * A.shape[1]), bottom, lbs[i], fontsize=11, \
                            ha='center', va='center', rotation=-45)
                    ax.text(i + ((0.3/10) * A.shape[1]), top, ubs[i], fontsize=11, \
                            ha='center', va='center', rotation=45)
            else:
                ax.set_xlim(x[0], x[-1])
        else:
            if len(ax.get_xticklabels()) < len(x):
                ax.set_xticks(x)
                ax.set_xticklabels(xtick_labels)
            if not show_bounds:
                xl, xr = ax.get_xlim()
                xl = x[0] if x[0] <= xl else xl
                xr = x[-1] if x[-1] >= xr else xr
                ax.set_xlim(xl, xr)

    if not show_bounds or is_xticklabels_off(ax):
        ax.tick_params(axis='x', labelsize=12)
        ax.tick_params(axis='y', labelsize=12) 
    
    # where to put the legend
    if labels is not None:
        ax.legend(loc="upper right") 
       
    # colorbar?
    if colorbar and isinstance(colorbar, tuple) and len(colorbar) >= 2 \
            and isinstance(colorbar[0], np.ndarray) and isinstance(colorbar[1], np.ndarray):
        vmin,vmax = 0.0, 1.0
        cbc, cbg = colorbar[0], colorbar[1]
        cbl = colorbar[2] if len(colorbar) > 2 and colorbar[2] else None
        Id = np.column_stack((cbg,cbc)).astype(object)
        Id = Id[np.argsort(Id[:, 0])] 
        c, g = Id[:,1:].astype(float), Id[:,0].astype(float)
        vmin, vmax = np.min(g), np.max(g)
        norm = mc.Normalize(vmin=vmin, vmax=vmax)
        cmap = ListedColormap(c)
        if cbl:
            ax.figure.colorbar(cm.ScalarMappable(norm=norm, cmap=cmap), \
                    orientation='vertical', label=cbl, pad=0.01, shrink=0.99)
        else:
            ax.figure.colorbar(cm.ScalarMappable(norm=norm, cmap=cmap), \
                        orientation='vertical', pad=0.01, shrink=0.99)

    # title?
    ax.set_title(title)

    return ax



"""scatter.py -- A customized scatter plotting module. 

    This module provides a customized scatter plotting functions for
    high-dimensional Pareto-optimal fronts. It also provides different
    relevant parameters, tools and utilities.

    Copyright (C) 2016
    Computational Optimization and Innovation (COIN) Laboratory
    Department of Computer Science and Engineering
    Michigan State University
    428 S. Shaw Lane, Engineering Building
    East Lansing, MI 48824-1226, USA

.. moduleauthor:: AKM Khaled Talukder <talukde1@msu.edu>

"""

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import matplotlib.colors as mc
from matplotlib.colors import ListedColormap
from mpl_toolkits.mplot3d import Axes3D
from viz.plotting.utils import pop, group_labels_by_appearance


__all__ = ["camera_angles", "plot", "camera_angles"]


# Some good camera angles for scatter plots.
camera_angles = {
    'dtlz2': {'3d': (60,10), '4d':(105,15), '8d': (15,15)}, \
    'dtlz2-nbi': {'3d': (60,10), '4d':(105,15), '5d': (105,15), '8d': (110,15)}, \
    'debmdk': {'3d': (115,5), '4d': (105,15), '8d': (110,15)}, \
    'debmdk-nbi': {'3d': (115,5), '4d': (105,15), '8d': (110,15)}, \
    'debmdk-all': {'3d': (115,5), '4d': (105,15), '8d': (110,15)}, \
    'debmdk-all-nbi': {'3d': (115,5), '4d': (105,15), '8d': (110,15)}, \
    'dtlz8': {'3d': (110,15), '4d': (-65,15), '6d': (-65,15), '8d': (-65,15)}, \
    'dtlz8-nbi': {'3d': (110,15), '4d': (-65,15), '6d': (-65,15), '8d': (-65,15)}, \
    'c2dtlz2': {'3d': (30,10), '4d': (-75,15), '5d': (-65,15), '8d': (110,15)}, \
    'c2dtlz2-nbi': {'3d': (30,10), '4d': (-75,15), '5d': (-65,15), '8d': (110,15)}, \
    'cdebmdk': {'3d': (30,10), '4d': (-75,15), '8d': (110,15)}, \
    'cdebmdk-nbi': {'3d': (30,10), '4d': (-75,15), '8d': (110,15)}, \
    'c0dtlz2': {'3d': (30,10), '4d': (-75,15), '8d': (110,15)}, \
    'c0dtlz2-nbi': {'3d': (30,10), '4d': (-75,15), '8d': (110,15)}, \
    'crash-nbi': {'3d': (30,15)}, 'crash-c1-nbi': {'3d': (30,15)}, 'crash-c2-nbi': {'3d': (30,15)}, \
    'gaa': {'10d': (-65,15)}, \
    'gaa-nbi': {'10d': (-65,15)}
}


def plot(A, ax=None, s=1, c=mc.TABLEAU_COLORS['tab:blue'], **kwargs):
    r"""A scatter plot function.

    This uses `matplotlib.axes.Axes.scatter` function to do a scatter plot.

    Parameters
    ----------
    A : ndarray 
        `n` number of `m` dim. points to be plotted.
    ax : An `mpl_toolkits.mplot3d.axes.Axes3D` or an `matplotlib.axes.Axes` object, optional
        Axes to be used to plotting. Default `None` when optional.
    s : int or 1-D array_like, optional
        Point size, or an array of point sizes. Default 1 when optional.
    c : A `matplotlib.colors` object, str or an array RGBA color values.
        Colors to be used. Default `mc.TABLEAU_COLORS['tab:blue']` when optional.

    Other Parameters
    ----------------
    euler : tuple (i.e. a pair) of int, optional
        The azmiuth and elevation angle. Default `(-60,30)` when optional.
    title : str, optional
        The plot title. Default `None` when optional.
    axes : tuple of int, optional
        The list of columns of `A` to be plotted. Default `(0, 1, 2)` when optional.
    labels : str, array_like or list of str, optional
        A string or an array/list of strings for labeling each point. Which basically
        means the class label of each row. Default `None` when optional. This will be
        used to set the legend in the figure. If `None` there will be no legend.
    colorbar : (Cbc, Cbg, Cbl, Cbp) a tuple of (ndarray, ndarray, str, float), optional
        If a user wants to put a colorbar, a tuple `(Cbc, Cbg, Cbl)` tuple can be 
        provided. `Cbc` is an array of RGBA color values or an `matplotlib.colors` 
        object. The gradient of the colorbar is specified in `Cbg` which is an 1-D 
        array of float. Cbl is the label of the colorbar, a string. `Cbp` is the 
        colorbar padding width in `float`. `colorbar` is default `None` when optional.
    xlim : tuple (i.e. a pair) of int, optional
        The limits on the X-axis. Default `None` when optional.
    ylim : tuple (i.e. a pair) of int, optional 
        The limits on the Y-axis. Default `None` when optional.
    zlim : tuple (i.e. a pair) of int, optional 
        The limits on the Z-axis. Default `None` when optional.
    label_prefix : str, optional
        The axis-label-prefix to be used, default `r"$f_{{{:d}}}$"` when optional.
    label_fontsize : str or int, optional
        The fontsize for the axes labels. Default `'large'` when optional.
    **kwargs : dict
        All other keyword args for matplotlib's `scatter()` function.

    Returns
    -------
    ax : An `mpl_toolkits.mplot3d.axes.Axes3D` or `matplotlib.axes.Axes` object
        An `mpl_toolkits.mplot3d.axes.Axes3D` or an `matplotlib.axes.Axes` object.
    """

    # all other parameters
    euler = kwargs['euler'] if kwargs and 'euler' in kwargs else (-60, 30)
    title = kwargs['title'] if kwargs and 'title' in kwargs else None
    axes = kwargs['axes'] if kwargs and 'axes' in kwargs else (0, 1, 2)
    labels = kwargs['labels'] if kwargs and 'labels' in kwargs else None
    colorbar = kwargs['colorbar'] if kwargs and 'colorbar' in kwargs else None
    xlim = kwargs['xlim'] if kwargs and 'xlim' in kwargs else None
    ylim = kwargs['ylim'] if kwargs and 'ylim' in kwargs else None
    zlim = kwargs['zlim'] if kwargs and 'zlim' in kwargs else None
    label_prefix = kwargs['label_prefix'] if kwargs and 'label_prefix' in kwargs else r"$f_{{{:d}}}$"
    label_fontsize = kwargs['label_fontsize'] if kwargs and 'label_fontsize' in kwargs else 'large'
            
    # remove once they are read
    kwargs = pop(kwargs, 'euler')
    kwargs = pop(kwargs, 'title')
    kwargs = pop(kwargs, 'axes')
    kwargs = pop(kwargs, 'labels')
    kwargs = pop(kwargs, 'colorbar')
    kwargs = pop(kwargs, 'xlim')
    kwargs = pop(kwargs, 'ylim')
    kwargs = pop(kwargs, 'zlim')
    kwargs = pop(kwargs, 'label_prefix')
    kwargs = pop(kwargs, 'label_fontsize')

    # decide on what kind of axes to use
    if not ax:
        ax = Axes3D(plt.figure()) if A.shape[1] > 2 else plt.figure().gca()

    if ax:
        # do the plot
        if labels is not None:
            if isinstance(labels, str):
                if A.shape[1] > 2:
                    ax.scatter(A[:,axes[0]], A[:,axes[1]], A[:,axes[2]], s=s, c=c, label=labels, **kwargs)
                else:
                    ax.scatter(A[:,axes[0]], A[:,axes[1]], s=s, c=c, label=labels, **kwargs)
            else:
                if isinstance(labels, np.ndarray): 
                    labels = labels.tolist()
                label_groups = group_labels_by_appearance(labels)
                for i,v in enumerate(label_groups):
                    if A.shape[1] > 2:
                        ax.scatter(A[v[0],axes[0]], A[v[0],axes[1]], A[v[0],axes[2]], \
                                s=s[v[0]], c=c[v[0]], label=v[1], zorder=label_groups.shape[0]-i, **kwargs)
                    else:
                        ax.scatter(A[v[0],axes[0]], A[v[0],axes[1]], s=s[v[0]], c=c[v[0]], \
                                label=v[1], zorder=label_groups.shape[0]-i, **kwargs)
        else:
            if A.shape[1] > 2:
                ax.scatter(A[:,axes[0]], A[:,axes[1]], A[:,axes[2]], s=s, c=c, **kwargs)
            else:
                ax.scatter(A[:,axes[0]], A[:,axes[1]], s=s, c=c, **kwargs)

        # set limits, put labels, fix labels, view and title
        if A.shape[1] > 2:
            ax.set_xlim(ax.get_xlim() if xlim is None else xlim)
            ax.set_ylim(ax.get_ylim() if ylim is None else ylim)
            ax.set_zlim(ax.get_zlim() if zlim is None else zlim)
            if len(axes) > 0:
                ax.set_xlabel(label_prefix.format(axes[0] + 1), fontsize=label_fontsize)
            if len(axes) > 1:
                ax.set_ylabel(label_prefix.format(axes[1] + 1), fontsize=label_fontsize)
            if len(axes) > 2:
                ax.set_zlabel(label_prefix.format(axes[2] + 1), fontsize=label_fontsize)
            ax.xaxis.set_rotate_label(False)
            ax.yaxis.set_rotate_label(False)
            ax.zaxis.set_rotate_label(False)
            ax.view_init(euler[1], euler[0])
            ax.set_title(title, pad=0.1)
        else:
            ax.set_xlim(ax.get_xlim() if xlim is None else xlim)
            ax.set_ylim(ax.get_ylim() if ylim is None else ylim)
            if len(axes) > 0:
                ax.set_xlabel(label_prefix.format(axes[0] + 1), fontsize=label_fontsize)
            if len(axes) > 1:
                ax.set_ylabel(label_prefix.format(axes[1] + 1), fontsize=label_fontsize)
            ax.set_title(title, y=ax.get_ylim()[-1]-0.05)
        
        # colorbar?
        if colorbar and isinstance(colorbar, tuple) and len(colorbar) >= 2 \
                and isinstance(colorbar[0], np.ndarray) and isinstance(colorbar[1], np.ndarray):
            vmin,vmax = 0.0, 1.0
            cbc, cbg = colorbar[0], colorbar[1]
            cbl = colorbar[2] if len(colorbar) > 2 and colorbar[2] else None
            cbp = colorbar[3] if len(colorbar) > 3 and colorbar[3] else 0.01
            Id = np.column_stack((cbg,cbc)).astype(object)
            Id = Id[np.argsort(Id[:, 0])] 
            c, g = Id[:,1:].astype(float), Id[:,0].astype(float)
            vmin, vmax = np.min(g), np.max(g)
            norm = mc.Normalize(vmin=vmin, vmax=vmax)
            cmap = ListedColormap(c)
            if cbl:
                ax.figure.colorbar(cm.ScalarMappable(norm=norm, cmap=cmap), \
                        orientation='vertical', label=cbl, pad=cbp, shrink=0.5)
            else:
                ax.figure.colorbar(cm.ScalarMappable(norm=norm, cmap=cmap), \
                            orientation='vertical', pad=cbp, shrink=0.5) 

        # where to put the legend
        if labels:
            ax.legend(loc="best", ncol=1)

        return ax
    else:
        raise TypeError("A valid `mpl_toolkits.mplot3d.axes.Axes3D`/`matplotlib.axes.Axes` " 
                + "object is not found.")
